# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.packets.packet_config import PacketConfig
from openc3.packets.parsers.packet_item_parser import PacketItemParser
from openc3.tools.table_manager.table_item import TableItem
from openc3.utilities.logger import Logger

class TableItemParser(PacketItemParser):
    """
    This class parses the PARAMETER, APPEND_PARAMETER, etc. keywords and creates the
    associated TableItem objects in the specified table.
    """

    @classmethod
    def parse(cls, parser, table, warnings):
        """
        Create a new TableItemParser and parse the packet item

        Args:
            parser: A ConfigParser
            table: The table the item should be added to
            warnings: An array of warning strings

        Returns:
            The created TableItem
        """
        parser = cls(parser, warnings)
        parser.verify_parameters(PacketConfig.COMMAND)
        return parser.create_table_item(table)

    def create_table_item(self, table):
        """
        Parse the next table item

        Returns:
            A TableItem
        """
        try:
            item_name = self.parser.parameters[0].upper()
            if item_name in table.items:
                msg = f"Table {table.table_name} item {item_name} redefined."
                Logger.warn(msg)
                self.warnings.append(msg)

            if table.type == 'ROW_COLUMN':
                item_name = f"{item_name}0"
                table.num_columns += 1

            item = TableItem(
                item_name,
                self._get_bit_offset(),
                self._get_bit_size(),
                self._get_data_type(),
                self._get_endianness(table),
                self._get_array_size(),
                "ERROR",  # overflow
            )

            item.minimum = self._get_minimum()
            item.maximum = self._get_maximum()
            item.default = self._get_default()
            item.id_value = self._get_id_value(item)
            item.description = self._get_description()

            if self._append():
                item = table.append(item)
            else:
                item = table.define(item)
            return item
        except Exception as error:
            raise self.parser.error(error, self.usage)


# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.interfaces.stream_interface import StreamInterface
from openc3.streams.tcpip_client_stream import TcpipClientStream
from openc3.config.config_parser import ConfigParser


# Base class for interfaces that act as a TCP/IP client
class TcpipClientInterface(StreamInterface):
    # self.param hostname [String] Machine to connect to
    # self.param write_port [Integer] Port to write commands to
    # self.param read_port [Integer] Port to read telemetry from
    # self.param write_timeout [Float] Seconds to wait before aborting writes
    # self.param read_timeout [Float|None] Seconds to wait before aborting reads.
    #   Pass None to block until the read is complete.
    # self.param protocol_type [String] Name of the protocol to use
    #   with this interface
    # self.param protocol_args [Array<String>] Arguments to pass to the protocol
    def __init__(
        self,
        hostname,
        write_port,
        read_port,
        write_timeout,
        read_timeout,
        protocol_type=None,
        *protocol_args,
    ):
        super().__init__(protocol_type, protocol_args)
        self.hostname = hostname
        self.write_port = ConfigParser.handle_none(write_port)
        self.read_port = ConfigParser.handle_none(read_port)
        self.write_timeout = write_timeout
        self.read_timeout = read_timeout
        if not self.read_port:
            self.read_allowed = False
        if not self.write_port:
            self.write_allowed = False
        if not self.write_port:
            self.write_raw_allowed = False

    def connection_string(self):
        # Probably most common is write == read so handle that
        if self.write_port == self.read_port:
            return f"{self.hostname}:{self.write_port} (R/W)"

        result = ""
        if self.write_port:
            result += f"{self.hostname}:{self.write_port} (write)"
        if self.read_port:
            if len(result) != 0:
                result += " "
            result += f"{self.hostname}:{self.read_port} (read)"
        return result

    # Connects the {TcpipClientStream} by passing the
    # initialization parameters to the {TcpipClientStream}.
    def connect(self):
        self.stream = TcpipClientStream(
            self.hostname,
            self.write_port,
            self.read_port,
            self.write_timeout,
            self.read_timeout,
        )
        super().connect()



# Copyright 2023 OpenC3, Inc
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.:

from openc3.api import WHITELIST
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.models.setting_model import SettingModel
from openc3.utilities.local_mode import LocalMode

WHITELIST.extend(
    [
        "list_settings",
        "get_all_settings",
        "get_setting",
        "get_settings",
        "set_setting",
        "save_setting",  # DEPRECATED
    ]
)


def list_settings(scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    return SettingModel.names(scope=scope)


def get_all_settings(scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    return SettingModel.all(scope=scope)


def get_setting(name, scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    setting = SettingModel.get(name=name, scope=scope)
    if setting:
        return setting["data"]
    else:
        return None


def get_settings(*settings, scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    result = []
    for name in settings:
        result.append(get_setting(name, scope))
    return result


def set_setting(name, data, local_mode=True, scope=OPENC3_SCOPE):
    authorize(permission="admin", scope=scope)
    SettingModel.set({"name": name, "data": data}, scope=scope)
    if local_mode:
        LocalMode.save_setting(scope, name, data)


# DEPRECATED
def save_setting(name, data, local_mode=True, scope=OPENC3_SCOPE):
    set_setting(name, data, local_mode, scope)



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.config.config_parser import ConfigParser
from openc3.interfaces.protocols.protocol import Protocol
from openc3.accessors.binary_accessor import BinaryAccessor
from openc3.packets.packet import Packet
from openc3.utilities.logger import Logger
from openc3.utilities.extract import hex_to_byte_string


# Reads all data available on the interface and creates a packet
# with that data.
class BurstProtocol(Protocol):
    # self.param discard_leading_bytes [Integer] The number of bytes to discard
    #   from the binary data after reading. Note that this is often
    #   used to remove a sync pattern from the final packet data.
    # self.param sync_pattern [String] String representing a hex number ("0x1234")
    #   that will be searched for in the raw data. Bytes encountered before
    #   this pattern is found are discarded.
    # self.param fill_fields [Boolean] Fill any required fields when writing packets
    # self.param allow_empty_data [True/False/None] See Protocol#initialize
    def __init__(
        self,
        discard_leading_bytes=0,
        sync_pattern=None,
        fill_fields=False,
        allow_empty_data=None,
    ):
        super().__init__(allow_empty_data) # Calls reset()
        self.discard_leading_bytes = int(discard_leading_bytes)
        self.sync_pattern = ConfigParser.handle_none(sync_pattern)
        if self.sync_pattern:
            self.sync_pattern = hex_to_byte_string(self.sync_pattern)
        self.fill_fields = ConfigParser.handle_true_false(fill_fields)

    def reset(self):
        super().reset()
        self.data = b""
        # self.data.force_encoding('ASCII-8BIT')
        self.sync_state = "SEARCHING"

    # Reads from the interface. It can look for a sync pattern before
    # creating a Packet. It can discard a set number of bytes at the try:ning
    # before creating the Packet.
    #
    # Note= On the first call to this from any interface read(), data will contain a blank
    # string. Blank string is an opportunity for protocols to return any queued up packets.
    # If they have no queued up packets, they should pass the blank string down to chained
    # protocols giving them the same opportunity.
    #
    # self.return [String|None] Data for a packet consisting of the bytes read
    def read_data(self, data, extra=None):
        self.data += data
        if not (len(data) == 0 and extra is None):
            # Maintain extra from last read read_data
            self.extra = extra

        while True:
            control = self.handle_sync_pattern()
            if control and len(data) > 0:  # Only return here if not blank string test
                return (control, extra)

            # Reduce the data to a single packet
            packet_data, extra = self.reduce_to_single_packet()
            if packet_data == "RESYNC":
                self.sync_state = "SEARCHING"
                if len(data) > 0:  # Only immediately resync if not blank string test
                    continue

            # Potentially allow blank string to be sent to other protocols if no packet is ready in this one
            if isinstance(packet_data, str):
                if (len(data) <= 0) and packet_data != "DISCONNECT":
                    # On blank string test, return blank string (if not we had a packet or need disconnect)
                    # The base class handles the special match of returning STOP if on the last protocol in the
                    # chain
                    return super().read_data(data, extra)
                else:
                    return (
                        packet_data,
                        extra,
                    )  # Return any control code if not on blank string test

            self.sync_state = "SEARCHING"

            # Discard leading bytes if necessary
            if self.discard_leading_bytes > 0:
                packet_data = packet_data[self.discard_leading_bytes :]
            return (packet_data, extra)

    # Called to perform modifications on a command packet before it is sent
    #
    # self.param packet [Packet] Original packet
    # self.return [Packet] Potentially modified packet
    def write_packet(self, packet: Packet):
        # If we're filling the sync pattern and the sync pattern is part of the
        # packet (since we're not discarding any leading bytes) then we have to
        # fill the sync pattern in the actual packet so do it here.
        if self.fill_fields and self.sync_pattern and self.discard_leading_bytes == 0:
            # Directly write the packet buffer and fill in the sync pattern
            BinaryAccessor.write(
                self.sync_pattern,
                0,
                len(self.sync_pattern) * 8,
                "BLOCK",
                packet.buffer_no_copy(),
                "BIG_ENDIAN",
                "ERROR",
            )
        return packet

    # Called to perform modifications on write data before sending it to the interface
    #
    # self.param data [String] Raw packet data
    # self.return [String] Potentially modified packet data
    def write_data(self, data, extra=None):
        # If we're filling the sync pattern and discarding the leading bytes
        # during a read then we need to put them back during a write.
        # If we're discarding the bytes then by definition they can't be part
        # of the packet so we just modify the data.
        if self.fill_fields and self.discard_leading_bytes > 0:
            data = bytearray(b"\x00" * self.discard_leading_bytes) + data
            if self.sync_pattern:
                BinaryAccessor.write(
                    self.sync_pattern,
                    0,
                    len(self.sync_pattern) * 8,
                    "BLOCK",
                    data,
                    "BIG_ENDIAN",
                    "ERROR",
                )
        return super().write_data(data, extra)

    # self.return [Boolean] control code (None, 'STOP')
    def handle_sync_pattern(self):
        if self.sync_pattern and self.sync_state == "SEARCHING":
            while True:
                # Make sure we have some data to look for a sync word in
                if len(self.data) < len(self.sync_pattern):
                    return "STOP"

                # Find the beginning of the sync pattern
                try:
                    sync_index = self.data.index(self.sync_pattern[0])
                    # Make sure we have enough data for the whole sync pattern past this index
                    if len(self.data) < (sync_index + len(self.sync_pattern)):
                        return "STOP"

                    # Check for the rest of the sync pattern
                    found = True
                    index = sync_index
                    for byte in self.sync_pattern:
                        if self.data[index] != byte:
                            found = False
                            break
                        index += 1

                    if found:
                        if sync_index != 0:
                            self.log_discard(sync_index, True)
                            # Delete Data Before Sync Pattern
                            self.data = self.data[sync_index:]
                        self.sync_state = "FOUND"
                        return None

                    else:  # not found
                        self.log_discard(sync_index, False)
                        # Delete Data Before and including first character of suspected sync Pattern
                        self.data = self.data[(sync_index + 1) :]
                        continue

                except ValueError:  # sync_index = None
                    self.log_discard(len(self.data), False)
                    self.data = b""
                    return "STOP"
        return None

    def log_discard(self, length, found):
        name = ""
        if self.interface:
            name = self.interface.name
        Logger.error(f"{name}: Sync {'not ' if not found else ''}found. Discarding {length} bytes of data.")
        pdata = self.data
        if len(self.data) < 6:
            pdata = self.data[:]
            pdata += b"\x00\x00\x00\x00\x00\x00"
        Logger.error(
            "Starting 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X\n"
            % (pdata[0], pdata[1], pdata[2], pdata[3], pdata[4], pdata[5])
        )

    def reduce_to_single_packet(self):
        if len(self.data) <= 0:
            # Need some data
            return ("STOP", self.extra)

        # Reduce to packet data and clear data for next packet
        packet_data = self.data[:]
        self.data = b""
        return (packet_data, self.extra)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.topics.topic import Topic
from openc3.system.system import System
from openc3.utilities.json import JsonEncoder
from openc3.environment import OPENC3_SCOPE


class RouterTopic(Topic):
    # Generate a list of topics for this router. This includes the router itself
    # and all the targets which are assigned to this router.
    @classmethod
    def topics(cls, router, scope=OPENC3_SCOPE):
        topics = []
        topics.append(f"{{{scope}__CMD}}ROUTER__{router.name}")
        for target_name in router.tlm_target_names:
            for _, packet in System.telemetry.packets(target_name).items():
                topics.append(f"{scope}__TELEMETRY__{{{packet.target_name}}}__{packet.packet_name}")
        return topics

    @classmethod
    def receive_telemetry(cls, router, scope=OPENC3_SCOPE):
        while True:
            for topic, msg_id, msg_hash, redis in Topic.read_topics(RouterTopic.topics(router, scope)):
                result = yield topic, msg_id, msg_hash, redis
                if "CMD}ROUTER" in topic:
                    ack_topic = topic.split("__")
                    ack_topic[1] = "ACK" + ack_topic[1]
                    ack_topic = "__".join(ack_topic)
                    Topic.write_topic(ack_topic, {"result": result}, msg_id, 100)

    @classmethod
    def route_command(cls, packet, target_names, scope=OPENC3_SCOPE):
        if packet.identified():
            topic = f"{{{scope}__CMD}}TARGET__{packet.target_name}"
            Topic.write_topic(
                topic,
                {
                    "target_name": packet.target_name,
                    "cmd_name": packet.packet_name,
                    "cmd_buffer": json.dumps(packet.buffer_no_copy(), cls=JsonEncoder),
                },
                "*",
                100,
            )
        elif len(target_names) == 1:
            topic = f"{{{scope}__CMD}}TARGET__{target_names[0]}"
            target_name = "UNKNOWN"
            if packet.target_name is not None:
                target_name = packet.target_name
            Topic.write_topic(
                topic,
                {
                    "target_name": target_name,
                    "cmd_name": "UNKNOWN",
                    "cmd_buffer": json.dumps(packet.buffer_no_copy(), cls=JsonEncoder),
                },
                "*",
                100,
            )
        else:
            target_name = "UNKNOWN"
            if packet.target_name is not None:
                target_name = packet.target_name
            packet_name = "UNKNOWN"
            if packet.packet_name is not None:
                packet = packet.packet_name
            raise RuntimeError(f"No route for command: {target_name} {packet_name}")

    @classmethod
    def connect_router(cls, router_name, *router_params, scope=OPENC3_SCOPE):
        if router_params and len(router_params) == 0:
            Topic.write_topic(
                f"{{{scope}__CMD}}ROUTER__{router_name}",
                {"connect": "True", "params": json.dumps(router_params)},
                "*",
                100,
            )
        else:
            Topic.write_topic(f"{{{scope}__CMD}}ROUTER__{router_name}", {"connect": "True"}, "*", 100)

    @classmethod
    def disconnect_router(cls, router_name, scope=OPENC3_SCOPE):
        Topic.write_topic(f"{{{scope}__CMD}}ROUTER__{router_name}", {"disconnect": "True"}, "*", 100)

    @classmethod
    def start_raw_logging(cls, router_name, scope=OPENC3_SCOPE):
        Topic.write_topic(f"{{{scope}__CMD}}ROUTER__{router_name}", {"log_stream": "True"}, "*", 100)

    @classmethod
    def stop_raw_logging(cls, router_name, scope=OPENC3_SCOPE):
        Topic.write_topic(f"{{{scope}__CMD}}ROUTER__{router_name}", {"log_stream": "False"}, "*", 100)

    @classmethod
    def shutdown(cls, router, scope=OPENC3_SCOPE):
        Topic.write_topic(f"{{{scope}__CMD}}ROUTER__{router.name}", {"shutdown": "True"}, "*", 100)

    @classmethod
    def router_cmd(cls, router_name, cmd_name, *cmd_params, scope=OPENC3_SCOPE):
        data = {}
        data["cmd_name"] = cmd_name
        data["cmd_params"] = cmd_params
        Topic.write_topic(
            f"{{{scope}__CMD}}ROUTER__{router_name}",
            {"router_cmd": json.dumps(data)},
            "*",
            100,
        )

    @classmethod
    def protocol_cmd(
        cls,
        router_name,
        cmd_name,
        *cmd_params,
        read_write="READ_WRITE",
        index=-1,
        scope=OPENC3_SCOPE,
    ):
        data = {}
        data["cmd_name"] = cmd_name
        data["cmd_params"] = cmd_params
        data["read_write"] = str(read_write).upper()
        data["index"] = index
        Topic.write_topic(
            f"{{{scope}__CMD}}ROUTER__{router_name}",
            {"protocol_cmd": json.dumps(data)},
            "*",
            100,
        )



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import time
import json
from openc3.topics.topic import Topic
from openc3.utilities.store_queued import EphemeralStoreQueued
from openc3.top_level import HazardousError, CriticalCmdError
from openc3.utilities.time import to_nsec_from_epoch
from openc3.utilities.json import JsonEncoder


class CommandTopic(Topic):
    COMMAND_ACK_TIMEOUT_S = 30

    @classmethod
    def write_packet(cls, packet, scope):
        topic = f"{scope}__COMMAND__{{{packet.target_name}}}__{packet.packet_name}"
        msg_hash = {
            "time": to_nsec_from_epoch(packet.packet_time),
            "received_time": to_nsec_from_epoch(packet.received_time),
            "target_name": packet.target_name,
            "packet_name": packet.packet_name,
            "received_count": packet.received_count,
            "stored": str(packet.stored),
            "buffer": bytes(packet.buffer_no_copy()),
        }
        EphemeralStoreQueued.write_topic(topic, msg_hash)

    @classmethod
    def send_command(cls, command, timeout, scope):
        if timeout is None:
            timeout = cls.COMMAND_ACK_TIMEOUT_S
        ack_topic = f"{{{scope}__ACKCMD}}TARGET__{command['target_name']}"
        Topic.update_topic_offsets([ack_topic])
        # Save the existing cmd_params Hash and JSON generate before writing to the topic
        cmd_params = command["cmd_params"]
        command["cmd_params"] = json.dumps(command["cmd_params"], cls=JsonEncoder)
        cmd_id = Topic.write_topic(
            f"{{{scope}__CMD}}TARGET__{command['target_name']}",
            command,
            "*",
            100,
        )
        start_time = time.time()
        while (time.time() - start_time) < timeout:
            for _, _, msg_hash, _ in Topic.read_topics([ack_topic]):
                if msg_hash[b"id"] == cmd_id:
                    result = msg_hash[b"result"].decode()
                    if result == "SUCCESS":
                        return command["target_name"], command["cmd_name"], cmd_params
                    # Check for HazardousError which is a special case
                    elif "HazardousError" in result:
                        cls.raise_hazardous_error(
                            msg_hash,
                            command["target_name"],
                            command["cmd_name"],
                            cmd_params,
                        )
                    elif "CriticalCmdError" in result:
                        cls.raise_critical_cmd_error(
                            msg_hash,
                            command["username"],
                            command["target_name"],
                            command["cmd_name"],
                            cmd_params,
                            command["cmd_string"],
                        )
                    else:
                        raise RuntimeError(result)
        raise RuntimeError(f"Timeout of {timeout}s waiting for cmd ack")

    ###########################################################################
    # PRIVATE implementation details
    ###########################################################################

    @classmethod
    def raise_hazardous_error(cls, msg_hash, target_name, cmd_name, cmd_params):
        _, description, formatted = msg_hash[b"result"].decode().split("\n")
        # Create and populate a new HazardousError and raise it up
        # The _cmd method in script/commands.rb rescues this and calls prompt_for_hazardous
        error = HazardousError()
        error.target_name = target_name
        error.cmd_name = cmd_name
        error.cmd_params = cmd_params
        error.hazardous_description = description
        error.formatted = formatted

        # No Logger.info because the error is already logged by the Logger.info "Ack Received ...
        raise error

    @classmethod
    def raise_critical_cmd_error(cls, msg_hash, username, target_name, cmd_name, cmd_params, cmd_string):
        _, uuid = msg_hash[b"result"].decode().split("\n")
        # Create and populate a new CriticalCmdError and raise it up
        # The _cmd method in script/commands.rb rescues this and calls prompt_for_critical_cmd
        error = CriticalCmdError()
        error.uuid = uuid
        error.username = username
        error.target_name = target_name
        error.cmd_name = cmd_name
        error.cmd_params = cmd_params
        error.cmd_string = cmd_string

        # No Logger.info because the error is already logged by the Logger.info "Ack Received ...
        raise error



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import re
import json
from openc3.topics.topic import Topic
from openc3.system.system import System
from openc3.utilities.store import Store
from openc3.utilities.json import JsonEncoder, JsonDecoder


# LimitsEventTopic keeps track of not only the <SCOPE>__openc3_limits_events topic
# but also the ancillary key value stores. The LIMITS_CHANGE event updates the
# <SCOPE>__current_limits key. The LIMITS_SET event updates the <SCOPE>__limits_sets.
# The LIMITS_SETTINGS event updates the <SCOPE>__current_limits_settings.
# While this isn't a clean separation of topics (streams) and models (key-value)
# it helps maintain consistency as the topic and model are linked.
class LimitsEventTopic(Topic):
    @classmethod
    def write(cls, event, scope):
        match event["type"]:
            case "LIMITS_CHANGE":
                # The current_limits hash keeps only the current limits state of items
                # It is used by the API to determine the overall limits state
                field = f"{event['target_name']}__{event['packet_name']}__{event['item_name']}"
                Store.hset(f"{scope}__current_limits", field, event["new_limits_state"])

            case "LIMITS_SETTINGS":
                # Limits updated in limits_api.rb to avoid circular reference to TargetModel
                if cls.sets(scope=scope).get(event["limits_set"], None) is None:
                    Store.hset(f"{scope}__limits_sets", event["limits_set"], "false")

                field = f"{event['target_name']}__{event['packet_name']}__{event['item_name']}"
                limits_settings = Store.hget(f"{scope}__current_limits_settings", field)
                if limits_settings:
                    limits_settings = json.loads(limits_settings)
                else:
                    limits_settings = {}
                limits = {}
                limits["red_low"] = event["red_low"]
                limits["yellow_low"] = event["yellow_low"]
                limits["yellow_high"] = event["yellow_high"]
                limits["red_high"] = event["red_high"]
                if event.get("green_low") and event.get("green_high"):
                    limits["green_low"] = event["green_low"]
                    limits["green_high"] = event["green_high"]
                limits_settings[event["limits_set"]] = limits
                if event.get("persistence"):
                    limits_settings["persistence_setting"] = event["persistence"]
                if event.get("enabled", None) is not None:
                    limits_settings["enabled"] = event["enabled"]
                Store.hset(
                    f"{scope}__current_limits_settings",
                    field,
                    json.dumps(limits_settings),
                )

            case "LIMITS_ENABLE_STATE":
                field = f"{event['target_name']}__{event['packet_name']}__{event['item_name']}"
                limits_settings = Store.hget(f"{scope}__current_limits_settings", field)
                if limits_settings:
                    limits_settings = json.loads(limits_settings)
                else:
                    limits_settings = {}
                limits_settings["enabled"] = event["enabled"]
                Store.hset(
                    f"{scope}__current_limits_settings",
                    field,
                    json.dumps(limits_settings),
                )

            case "LIMITS_SET":
                sets = cls.sets(scope=scope)
                if sets.get(event["set"]) is None:
                    raise RuntimeError(f"Set '{event['set']}' does not exist!")

                # Set all existing sets to "false"
                sets = dict.fromkeys(sets, "false")
                sets[event["set"]] = "true"  # Enable the requested set
                Store.hset(f"{scope}__limits_sets", mapping=sets)
            case _:
                raise RuntimeError(f"Invalid limits event type '{event['type']}'")

        Topic.write_topic(
            f"{scope}__openc3_limits_events",
            {"event": json.dumps(event, cls=JsonEncoder)},
            "*",
            1000,
        )

    # Remove the JSON encoding to return hashes directly
    @classmethod
    def read(cls, offset=None, count=100, scope=None):
        final_result = []
        topic = f"{scope}__openc3_limits_events"
        if offset is not None:
            for topic, msg_id, msg_hash, redis in Topic.read_topics([topic], [offset], None, count):
                # result = Topic.read_topics([topic], [offset], None, count)
                # if len(result) != 0:
                # result is a hash with the topic key followed by an array of results
                # This returns just the array of arrays [[offset, hash], [offset, hash], ...]
                final_result.append([msg_id, msg_hash])
        else:
            result = Topic.get_newest_message(topic)
            if result:
                final_result = [result]
        parsed_result = []
        for offset, hash in final_result:
            parsed_result.append([offset, json.loads(hash[b"event"], cls=JsonDecoder)])
        return parsed_result

    @classmethod
    def out_of_limits(cls, scope):
        out_of_limits = []
        limits = Store.hgetall(f"{scope}__current_limits")
        # decode the binary string keys to strings
        limits = {k.decode(): v.decode() for (k, v) in limits.items()}
        for item, limits_state in limits.items():
            if limits_state in [
                "RED",
                "RED_HIGH",
                "RED_LOW",
                "YELLOW",
                "YELLOW_HIGH",
                "YELLOW_LOW",
            ]:
                target_name, packet_name, item_name = item.split("__")
                out_of_limits.append([target_name, packet_name, item_name, limits_state])
        return out_of_limits

    # Returns all the limits sets as keys with the value 'true' or 'false'
    # where only the active set is 'true'
    #
    # @return [Hash{String => String}] Set name followed by 'true' if enabled else 'false'
    @classmethod
    def sets(cls, scope):
        sets = Store.hgetall(f"{scope}__limits_sets")
        # decode the binary string keys to strings
        return {k.decode(): v.decode() for (k, v) in sets.items()}

    @classmethod
    def current_set(cls, scope):
        sets = LimitsEventTopic.sets(scope=scope)
        # Lookup the key with a true value because there should only ever be one
        try:
            return list(sets.keys())[list(sets.values()).index("true")]
        except ValueError:
            return "DEFAULT"

    # Cleanups up the current_limits and current_limits_settings keys for
    # a target or target/packet combination
    @classmethod
    def delete(cls, target_name, packet_name=None, scope=None):
        limits = Store.hgetall(f"{scope}__current_limits")
        # decode the binary string keys to strings
        limits = {k.decode(): v for (k, v) in limits.items()}
        for item, _ in limits.items():
            if packet_name:
                if re.match(rf"^{target_name}__{packet_name}__", item):
                    Store.hdel(f"{scope}__current_limits", item)
            else:
                if re.match(rf"^{target_name}__", item):
                    Store.hdel(f"{scope}__current_limits", item)

        limits_settings = Store.hgetall(f"{scope}__current_limits_settings")
        # decode the binary string keys to strings
        limits_settings = {k.decode(): v for (k, v) in limits_settings.items()}
        for item, _ in limits_settings.items():
            if packet_name:
                if re.match(rf"^{target_name}__{packet_name}__", item):
                    Store.hdel(f"{scope}__current_limits_settings", item)
            else:
                if re.match(rf"^{target_name}__", item):
                    Store.hdel(f"{scope}__current_limits_settings", item)

    # Update the local System based on overall state
    @classmethod
    def sync_system(cls, scope):
        all_limits_settings = Store.hgetall(f"{scope}__current_limits_settings")
        # decode the binary string keys to strings
        all_limits_settings = {k.decode(): v for (k, v) in all_limits_settings.items()}
        telemetry = System.telemetry.all()
        for item, limits_settings in all_limits_settings.items():
            target_name, packet_name, item_name = item.split("__")
            target = telemetry.get(target_name, None)
            if target is not None:
                packet = target.get(packet_name, None)
                if packet is not None:
                    limits_settings = json.loads(limits_settings)
                    enabled = limits_settings.get("enabled", None)
                    persistence = limits_settings.get("persistence_setting", 1)
                    for limits_set, settings in limits_settings.items():
                        if not isinstance(settings, dict):
                            continue
                        System.limits.set(
                            target_name,
                            packet_name,
                            item_name,
                            settings["red_low"],
                            settings["yellow_low"],
                            settings["yellow_high"],
                            settings["red_high"],
                            settings.get("green_low", None),
                            settings.get("green_high", None),
                            str(limits_set),
                            persistence,
                            enabled,
                        )
                    if enabled is not None:
                        if enabled:
                            System.limits.enable(target_name, packet_name, item_name)
                        else:
                            System.limits.disable(target_name, packet_name, item_name)

    # Update the local system based on limits events
    @classmethod
    def sync_system_thread_body(cls, scope):
        telemetry = System.telemetry.all()
        topics = [f"{scope}__openc3_limits_events"]
        for _, _, event, _ in Topic.read_topics(topics, timeout_ms=None):
            event = json.loads(event[b"event"], cls=JsonDecoder)
            match event["type"]:
                case "LIMITS_CHANGE":
                    pass  # Ignore
                case "LIMITS_SETTINGS":
                    target_name = event["target_name"]
                    packet_name = event["packet_name"]
                    item_name = event["item_name"]
                    target = telemetry.get(target_name)
                    if target:
                        packet = target.get(packet_name)
                        if packet:
                            enabled = event.get("enabled", None)
                            persistence = event.get("persistence", 1)
                            System.limits.set(
                                target_name,
                                packet_name,
                                item_name,
                                event["red_low"],
                                event["yellow_low"],
                                event["yellow_high"],
                                event["red_high"],
                                event.get("green_low", None),
                                event.get("green_high", None),
                                event["limits_set"],
                                persistence,
                                enabled,
                            )

                case "LIMITS_ENABLE_STATE":
                    target_name = event["target_name"]
                    packet_name = event["packet_name"]
                    item_name = event["item_name"]
                    target = telemetry.get(target_name)
                    if target:
                        packet = target.get(packet_name)
                        if packet:
                            enabled = event.get("enabled", False)
                            if enabled:
                                System.limits.enable(target_name, packet_name, item_name)
                            else:
                                System.limits.disable(target_name, packet_name, item_name)

                case "LIMITS_SET":
                    pass  # Ignore, System.limits_set() always queries Redis



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.models.model import Model


# Stores the status about an interface. This class also implements logic
# to handle status for a router since the functionality is identical
# (only difference is the Redis key used).
class InterfaceStatusModel(Model):
    INTERFACES_PRIMARY_KEY = "openc3_interface_status"
    ROUTERS_PRIMARY_KEY = "openc3_router_status"

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name: str, scope: str):
        return super().get(f"{scope}__{cls._get_key()}", name=name)

    @classmethod
    def names(cls, scope: str):
        return super().names(f"{scope}__{cls._get_key()}")

    @classmethod
    def all(cls, scope: str):
        return super().all(f"{scope}__{cls._get_key()}")

    # END NOTE

    # Helper method to return the correct type based on class name
    @classmethod
    def _get_type(cls):
        return cls.__name__.split("Model")[0].upper()

    # Helper method to return the correct primary key based on class name
    @classmethod
    def _get_key(cls):
        type_ = cls._get_type()
        match type_:
            case "INTERFACESTATUS":
                return InterfaceStatusModel.INTERFACES_PRIMARY_KEY
            case "ROUTERSTATUS":
                return InterfaceStatusModel.ROUTERS_PRIMARY_KEY
            case _:
                raise RuntimeError(f"Unknown type {type_} from class {cls.__name__}")

    def __init__(
        self,
        name,
        state,
        clients=0,
        txsize=0,
        rxsize=0,
        txbytes=0,
        rxbytes=0,
        txcnt=0,
        rxcnt=0,
        updated_at=None,
        plugin=None,
        scope=None,
    ):
        if self.__class__._get_type() == "INTERFACESTATUS":
            super().__init__(
                f"{scope}__{InterfaceStatusModel.INTERFACES_PRIMARY_KEY}",
                name=name,
                updated_at=updated_at,
                plugin=plugin,
                scope=scope,
            )
        else:
            super().__init__(
                f"{scope}__{InterfaceStatusModel.ROUTERS_PRIMARY_KEY}",
                name=name,
                updated_at=updated_at,
                plugin=plugin,
                scope=scope,
            )
        self.state = state
        self.clients = clients
        self.txsize = txsize
        self.rxsize = rxsize
        self.txbytes = txbytes
        self.rxbytes = rxbytes
        self.txcnt = txcnt
        self.rxcnt = rxcnt

    def as_json(self):
        return {
            "name": self.name,
            "state": self.state,
            "clients": self.clients,
            "txsize": self.txsize,
            "rxsize": self.rxsize,
            "txbytes": self.txbytes,
            "rxbytes": self.rxbytes,
            "txcnt": self.txcnt,
            "rxcnt": self.rxcnt,
            "plugin": self.plugin,
            "updated_at": self.updated_at,
        }



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.interfaces.protocols.terminated_protocol import TerminatedProtocol

# This file implements the COBS protocol as here=
# https=//en.wikipedia.org/wiki/Consistent_Overhead_Byte_Stuffing
# http=//www.stuartcheshire.org/papers/COBSforToN.pdf

# COBS is a framing protocol and is therefore expected to be used for packet deliniation

# Usage in plugin.txt=
#
# INTERFACE ...
#   PROTOCOL READ_WRITE CobsProtocol


class CobsProtocol(TerminatedProtocol):
    # self.param allow_empty_data [True/False/None] See Protocol#initialize
    def __init__(self, allow_empty_data=None):
        strip_read_termination = True
        discard_leading_bytes = 0
        sync_pattern = None
        fill_fields = False  # Handled in write_data below

        super().__init__(
            "",  # Write termination handled in write_data below
            "00",
            strip_read_termination,
            discard_leading_bytes,
            sync_pattern,
            fill_fields,
            allow_empty_data,
        )

    def read_data(self, data, extra=None):
        data, extra = super().read_data(data, extra)
        if len(data) <= 0 or isinstance(data, str):
            return (data, extra)

        result_data = b""
        while len(data) > 1:
            # Read the offset to the next zero byte
            # Note: This may be off the end of the data. If so, the packet is over
            zero_offset = data[0]
            if zero_offset == 0xFF:  # No zeros in this segment
                result_data += data[1:255]
                data = data[255:]
            elif zero_offset <= 1:  # End of data or 1 zero
                result_data += b"\x00"
                data = data[1:]
            else:  # Mid range zero or end of packet
                result_data += data[1:zero_offset]
                data = data[zero_offset:]
                if len(data) >= 1:
                    result_data += b"\x00"

        return (result_data, extra)

    def write_data(self, data, extra=None):
        # Intentionally not calling super()

        need_insert = False
        result_data = b""
        while len(data) > 0:
            try:
                index = data.index(b"\x00")
                if index > 253:
                    result_data += b"\xFF"
                    result_data += data[0:254]
                    data = data[254:]
                    need_insert = False
                elif index <= 253:
                    result_data += (index + 1).to_bytes(1, byteorder="big")
                    if index >= 1:
                        result_data += data[0:index]
                    data = data[(index + 1) :]
                    need_insert = True
            except ValueError:  # index not found
                if len(data) >= 254:
                    result_data += b"\xFF"
                    result_data += data[0:254]
                    data = data[254:]
                    need_insert = False
                else:
                    result_data += (len(data) + 1).to_bytes(1, byteorder="big")
                    result_data += data
                    data = b""
                    need_insert = False

        # Handle a zero at the end of the packet
        if need_insert:
            result_data += b"\x01"

        # Terminate message with 0x00
        result_data += b"\x00"

        return (result_data, extra)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
import openc3.script
from openc3.environment import OPENC3_SCOPE


def list_timelines(scope=OPENC3_SCOPE):
    response = openc3.script.API_SERVER.request("get", "/openc3-api/timeline", scope=scope)
    return _handle_response(response, "Failed to list timelines")


def create_timeline(name, color=None, scope=OPENC3_SCOPE):
    data = {}
    data["name"] = name
    if color:
        data["color"] = color
    response = openc3.script.API_SERVER.request("post", "/openc3-api/timeline", data=data, json=True, scope=scope)
    return _handle_response(response, "Failed to create timeline")


def get_timeline(name, scope=OPENC3_SCOPE):
    response = openc3.script.API_SERVER.request("get", f"/openc3-api/timeline/{name}", scope=scope)
    return _handle_response(response, "Failed to get timeline")


def set_timeline_color(name, color, scope=OPENC3_SCOPE):
    post_data = {}
    post_data["color"] = color
    response = openc3.script.API_SERVER.request(
        "post",
        f"/openc3-api/timeline/{name}/color",
        data=post_data,
        json=True,
        scope=scope,
    )
    return _handle_response(response, "Failed to set timeline color")


def delete_timeline(name, force=False, scope=OPENC3_SCOPE):
    url = f"/openc3-api/timeline/{name}"
    if force:
        url += "?force=true"
    response = openc3.script.API_SERVER.request("delete", url, scope=scope)
    return _handle_response(response, "Failed to delete timeline")


def create_timeline_activity(name, kind, start, stop, data={}, scope=OPENC3_SCOPE):
    kind = kind.lower()
    kinds = ["command", "script", "reserve"]
    if kind not in kinds:
        raise RuntimeError(f"Unknown kind: {kind}. Must be one of {', '.join(kinds)}.")
    post_data = {}
    post_data["start"] = start.isoformat()
    post_data["stop"] = stop.isoformat()
    post_data["kind"] = kind
    post_data["data"] = data
    response = openc3.script.API_SERVER.request(
        "post",
        f"/openc3-api/timeline/{name}/activities",
        data=post_data,
        json=True,
        scope=scope,
    )
    return _handle_response(response, "Failed to create timeline activity")


def get_timeline_activity(name, start, uuid, scope=OPENC3_SCOPE):
    response = openc3.script.API_SERVER.request("get", f"/openc3-api/timeline/{name}/activity/{start}/{uuid}", scope=scope)
    return _handle_response(response, "Failed to get timeline activity")


def get_timeline_activities(name, start=None, stop=None, limit=None, scope=OPENC3_SCOPE):
    url = f"/openc3-api/timeline/{name}/activities"
    if start and stop:
        url += f"?start={start}&stop={stop}"
    if limit:
        url += f"?limit={limit}"
    response = openc3.script.API_SERVER.request("get", url, scope=scope)
    return _handle_response(response, "Failed to get timeline activities")


def delete_timeline_activity(name, start, uuid, scope=OPENC3_SCOPE):
    response = openc3.script.API_SERVER.request(
        "delete", f"/openc3-api/timeline/{name}/activity/{start}/{uuid}", scope=scope
    )
    return _handle_response(response, "Failed to delete timeline activity")


# Helper method to handle the response
def _handle_response(response, error_message):
    if response is None:
        return None
    if response.status_code >= 400:
        result = json.loads(response.text)
        raise RuntimeError(f"{error_message} due to {result['message']}")
    # Not sure why the response body is empty (on delete) but check for that
    if response.text is None or len(response.text) == 0:
        return None
    else:
        return json.loads(response.text)



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.utilities.store import StoreMeta, Store, EphemeralStore
from openc3.utilities.sleeper import Sleeper
import time
import threading
import queue
import atexit

# Updated from top_level to remove circular dependency


# Attempt to gracefully kill a thread
# @param owner Object that owns the thread and may have a graceful_kill method
# @param thread The thread to gracefully kill
# @param timeout Timeout in seconds to wait for it to die gracefully
def kill_thread(owner, thread, timeout=1.0):
    if thread:
        if owner and hasattr(owner, "graceful_kill"):
            if threading.current_thread() != thread:
                owner.graceful_kill()
                thread.join(timeout=timeout)


class StoreQueued(metaclass=StoreMeta):
    # Variable that holds the singleton instance
    my_instance = None

    # Mutex used to ensure that only one instance is created
    instance_mutex = threading.Lock()

    # Get the singleton instance
    @classmethod
    def instance(cls, update_interval=1):
        if cls.my_instance:
            return cls.my_instance

        with cls.instance_mutex:
            cls.my_instance = cls(update_interval)
            return cls.my_instance

    def __init__(self, update_interval):
        self.update_interval = update_interval
        self.store = self.store_instance()
        # Queue to hold the store requests
        self.store_queue = queue.Queue()
        # Sleeper used to delay update thread
        self.update_sleeper = Sleeper()

        # Use atexit to shutdown cleanly no matter how we die
        atexit.register(self.shutdown)

        # Thread used to call methods on the store
        self.update_thread = threading.Thread(target=self.store_thread_body, daemon=True)
        self.update_thread.start()

    def set_update_interval(self, interval):
        if interval < self.update_interval and interval > 0.0:
            self.update_interval = interval

    def process_queue(self):
        if not self.store_queue.empty():
            # Pipeline the requests to redis to improve performance
            with self.store.redis_pool.get():
                with self.store.redis_pool.pipelined():
                    while not self.store_queue.empty():
                        action = self.store_queue.get()
                        getattr(self.store, action[0])(*action[1], **action[2])

    def store_thread_body(self):
        while True:
            start_time = time.time()

            self.process_queue()

            # Only check whether to update at a set interval
            run_time = time.time() - start_time
            sleep_time = self.update_interval - run_time
            if sleep_time < 0:
                sleep_time = 0
            if self.update_sleeper.sleep(sleep_time):
                break

    def shutdown(self):
        if self.update_sleeper:
            self.update_sleeper.cancel()
        if self.update_thread:
            kill_thread(self, self.update_thread)
        self.update_thread = None
        # Drain the queue before shutdown
        self.process_queue()

    # Record the message for pipelining by the thread
    def __getattr__(self, func):
        def method(*args, **kwargs):
            return self.store_queue.put([func, args, kwargs])

        return method

    # Returns the store we're working with
    def store_instance(self):
        return Store.instance()

    def graceful_kill(self):
        # Do nothing
        pass


class EphemeralStoreQueued(StoreQueued):
    # Variable that holds the singleton instance
    my_instance = None

    def store_instance(self):
        return EphemeralStore.instance()



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.packets.parsers.packet_parser import PacketParser
from openc3.tools.table_manager.table import Table
from openc3.top_level import Logger

class TableParser(PacketParser):
    """Parses the TABLE keyword definition in table configuration files."""

    def __init__(self, parser):
        super().__init__(parser)
        self.usage = "TABLE <TABLE NAME> <ENDIANNESS: BIG_ENDIAN or LITTLE_ENDIAN> <TYPE: KEY_VALUE or ROW_COLUMN> <NUM ROWS IF TYPE = ROW_COLUMN> <DESCRIPTION (Optional)>"

    @classmethod
    def parse_table(cls, parser, tables, warnings):
        """
        Parse a table definition

        Args:
            parser: Configuration parser
            tables: Hash of the currently defined tables
            warnings: Any warning strings generated while parsing this command will be appended to this array

        Returns:
            The parsed table
        """
        parser_obj = cls(parser)
        parser_obj.verify_parameters()
        return parser_obj.create_table(tables, warnings)

    def verify_parameters(self):
        """Verify the correct number of arguments to the TABLE keyword"""
        self.usage = 'TABLE <TABLE NAME> <ENDIANNESS: BIG_ENDIAN/LITTLE_ENDIAN> <DISPLAY: KEY_VALUE/ROW_COLUMN> <ROW_COLUMN ROW COUNT> <DESCRIPTION (Optional)>'
        self.parser.verify_num_parameters(3, 5, self.usage)

    def create_table(self, tables, warnings):
        """
        Create a table from the parameters

        Args:
            tables: All tables defined in the configuration
            warnings: List of warnings to append to

        Returns:
            The created table
        """
        params = self.parser.parameters
        table_name = params[0].upper()
        endianness = params[1].upper()
        if endianness != "BIG_ENDIAN" and endianness != "LITTLE_ENDIAN":
            raise self.parser.error(
                f"Invalid endianness {params[1]}. Must be BIG_ENDIAN or LITTLE_ENDIAN.",
                self.usage,
            )

        type_param = params[2].upper()
        if type_param in ("KEY_VALUE", "ONE_DIMENSIONAL"):  # :ONE_DIMENSIONAL is deprecated
            self.parser.verify_num_parameters(3, 4, self.usage)
            description = params[3] if len(params) > 3 else ""
            num_rows = None
        elif type_param in ("ROW_COLUMN", "TWO_DIMENSIONAL"):  # :TWO_DIMENSIONAL is deprecated
            self.parser.verify_num_parameters(4, 5, self.usage)
            num_rows = int(params[3])
            description = params[4] if len(params) > 4 else ""
        else:
            raise self.parser.error(
                f"Invalid display type {params[2]}. Must be KEY_VALUE or ROW_COLUMN.",
                self.usage,
            )

        table = Table(table_name, endianness, type_param, description, self.parser.filename)
        if type_param in ("ROW_COLUMN", "TWO_DIMENSIONAL"):
            table.num_rows = num_rows

        return self.finish_create_table(table, tables, warnings)

    @classmethod
    def check_for_duplicate(cls, tables, table):
        """
        Check for duplicate table definition

        Args:
            tables: All tables defined in the configuration
            table: The table to check

        Returns:
            Warning message or None
        """
        msg = None
        if table.TARGET in tables and table.table_name in tables[table.TARGET]:
            msg = f"Table {table.table_name} redefined."
            Logger.warn(msg)
        return msg

    @classmethod
    def finish_create_table(cls, table, tables, warnings):
        """
        Finish table creation by checking for duplicates

        Args:
            table: The table to finish
            tables: All tables defined in the configuration
            warnings: List of warnings to append to

        Returns:
            The finished table
        """
        warning = cls.check_for_duplicate(tables, table)
        if warning:
            warnings.append(warning)
        return table


# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import sys
import copy
import re
from io import StringIO
from openc3.script.exceptions import StopScript, SkipScript
import openc3.script
from openc3.io.stdout import Stdout
from openc3.io.stderr import Stderr


class Suite:
    """Base class for Script Runner suites. OpenC3 Suites inherit from Suite
    and can implement setup and teardown methods. Script groups are added via add_group(Group)
    and individual scripts added via add_script(Group, script_method)."""

    def scripts(self):
        if not hasattr(self, "_scripts"):
            self._scripts = {}
        return self._scripts

    def plans(self):
        if not hasattr(self, "_plans"):
            self._plans = []
        return self._plans

    ###########################################################################
    # START PUBLIC API
    ###########################################################################

    # Explicitly avoid creating an initialize method which forces users to call super()

    # Add a group to the suite
    def add_group(self, group_class):
        if not issubclass(group_class, Group):
            raise RuntimeError(
                f"add_group received with {group_class}({group_class.__class__}) but must subclass Group"
            )
        if not self.scripts().get(group_class, None):
            self.scripts()[group_class] = group_class()
        self.plans().append(["GROUP", group_class, None])

    # Add a script to the suite
    def add_script(self, group_class, script):
        if not issubclass(group_class, Group):
            raise RuntimeError(
                f"add_script received with {group_class}({group_class.__class__}) but must subclass Group"
            )
        if not self.scripts().get(group_class, None):
            self.scripts()[group_class] = group_class()
        self.plans().append(["SCRIPT", group_class, script])

    # Add a group setup to the suite
    def add_group_setup(self, group_class):
        if not issubclass(group_class, Group):
            raise RuntimeError(
                f"add_group_setup received with {group_class}({group_class.__class__}) but must subclass Group"
            )
        if not self.scripts().get(group_class, None):
            self.scripts()[group_class] = group_class()
        self.plans().append(["GROUP_SETUP", group_class, None])

    # Add a group teardown to the suite
    def add_group_teardown(self, group_class):
        if not issubclass(group_class, Group):
            raise RuntimeError(
                f"add_group_teardown received with {group_class}({group_class.__class__}) but must subclass Group"
            )
        if not self.scripts().get(group_class, None):
            self.scripts()[group_class] = group_class()
        self.plans().append(["GROUP_TEARDOWN", group_class, None])

    ###########################################################################
    # END PUBLIC API
    ###########################################################################

    def __eq__(self, other):
        return self.name == other.name

    # Name of the suite
    def name(self):
        if self.__class__ != Suite:
            return self.__class__.__name__
        else:
            return "UnassignedSuite"

    # Returns the number of scripts in the suite including setup and teardown methods
    def get_num_scripts(self):
        num_scripts = 0
        for type, group_class, _ in self.plans():
            if type == "GROUP":
                num_scripts += group_class.get_num_scripts()
            else:
                num_scripts += 1

        if hasattr(self, "setup"):
            num_scripts += 1
        if hasattr(self, "teardown"):
            num_scripts += 1
        return num_scripts

    # Run all the scripts
    def run(self):
        ScriptResult.suite = self.name()
        ScriptStatus.instance().total = self.get_num_scripts()
        results = []

        # Setup the suite
        result = self.run_setup(True)
        if result:
            results.append(result)
            yield result
            if result.stopped:
                raise StopScript

        # Run each script
        for type, group_class, script in self.plans():
            match type:
                case "GROUP":
                    for result in self.run_group(group_class, True):
                        results.append(result)
                        yield result
                        if result.stopped:
                            raise StopScript
                case "SCRIPT":
                    result = self.run_script(group_class, script, True)
                    results.append(result)
                    yield result
                    if (result.exceptions and group_class.abort_on_exception) or result.stopped:
                        raise StopScript
                case "GROUP_SETUP":
                    result = self.run_group_setup(group_class, True)
                    if result:
                        results.append(result)
                        yield result
                        if (result.exceptions and group_class.abort_on_exception) or result.stopped:
                            raise StopScript

                case "GROUP_TEARDOWN":
                    result = self.run_group_teardown(group_class, True)
                    if result:
                        results.append(result)
                        yield result
                        if (result.exceptions and group_class.abort_on_exception) or result.stopped:
                            raise StopScript

        # Teardown the suite
        result = self.run_teardown(True)
        if result:
            results.append(result)
            yield result
            if result.stopped:
                raise StopScript
        ScriptResult.suite = None
        return results

    # Run a specific group
    def run_group(self, group_class, internal=False):
        results = []
        if not internal:
            ScriptResult.suite = self.name()

        # Determine if this group_class is in the plan and the number of scripts associated with this group_class
        in_plan = False
        num_scripts = 0
        for plan_type, plan_group_class, plan_script in self.plans():
            if plan_type == "GROUP" and group_class == plan_group_class:
                in_plan = True

            if (
                (plan_type == "GROUP_SETUP" and group_class == plan_group_class)
                or (plan_type == "GROUP_TEARDOWN" and group_class == plan_group_class)
                or (plan_script and group_class == plan_group_class)
            ):
                num_scripts += 1

        if in_plan:
            if not internal:
                ScriptStatus.instance().total = group_class.get_num_scripts()
            for result in self.scripts()[group_class].run():
                results.append(result)
                yield result
                if result.stopped:
                    raise StopScript
        else:
            if not internal:
                ScriptStatus.instance().total = num_scripts

            # Run each setup, teardown, or script associated with this group_class in the order
            # defined in the plan
            for plan_type, plan_group_class, plan_script in self.plans():
                if plan_group_class == group_class:
                    match plan_type:
                        case "SCRIPT":
                            result = self.run_script(plan_group_class, plan_script, True)
                            results.append(result)
                            yield result
                        case "GROUP_SETUP":
                            result = self.run_group_setup(plan_group_class, True)
                            if result:
                                results.append(result)
                                yield result

                        case "GROUP_TEARDOWN":
                            result = self.run_group_teardown(plan_group_class, True)
                            if result:
                                results.append(result)
                                yield result
        if not internal:
            ScriptResult.suite = None
        return results

    # Run a specific script
    def run_script(self, group_class, script, internal=False):
        if not internal:
            ScriptResult.suite = self.name()
        if not internal:
            ScriptStatus.instance().total = 1
        result = self.scripts()[group_class].run_script(script)
        if not internal:
            ScriptResult.suite = None
        return result

    def run_setup(self, internal=False):
        if not internal:
            ScriptResult.suite = self.name()
        result = None
        if "setup" in dir(self) and len(self.scripts()) > 0:
            if not internal:
                ScriptStatus.instance().total = 1
            ScriptStatus.instance().status = f"{self.__class__.__name__} : setup"
            # Get the first group
            result = next(iter(self.scripts().values())).run_method(self, "setup")

        if not internal:
            ScriptResult.suite = None
        return result

    def run_teardown(self, internal=False):
        if not internal:
            ScriptResult.suite = self.name()
        result = None
        if "teardown" in dir(self) and len(self.scripts()) > 0:
            if not internal:
                ScriptStatus.instance().total = 1
            ScriptStatus.instance().status = f"{self.__class__} : teardown"
            # Get the first group
            result = next(iter(self.scripts().values())).run_method(self, "teardown")

        if not internal:
            ScriptResult.suite = None
        return result

    def run_group_setup(self, group_class, internal=False):
        if not internal:
            ScriptResult.suite = self.name()
        if not internal:
            ScriptStatus.instance().total = 1
        result = self.scripts()[group_class].run_setup()
        if not internal:
            ScriptResult.suite = None
        return result

    def run_group_teardown(self, group_class, internal=False):
        if not internal:
            ScriptResult.suite = self.name()
        if not internal:
            ScriptStatus.instance().total = 1
        result = self.scripts()[group_class].run_teardown()
        if not internal:
            ScriptResult.suite = None
        return result


class Group:
    """Base class for a group. All OpenC3 Script Runner scripts should inherit Group
    and then implement scripts methods starting with 'script_', 'test_', or 'op_'
    e.g. script_mech_open, test_mech_open, op_mech_open."""

    abort_on_exception = False
    current_result = None

    # Explicitly avoid creating an initialize method which forces users to call super()

    @classmethod
    def scripts(cls):
        # Find all the script methods
        result = [
            func
            for func in dir(cls)
            if callable(getattr(cls, func))
            and re.search(r"^test|^script|op_", func)
            and func != "scripts"
            and func != "test_cases"
        ]
        result.sort()
        return result

    # Name of the script group
    def name(self):
        if self.__class__ != Group:
            return self.__class__.__name__
        else:
            return "UnnamedGroup"

    # Run all the scripts
    def run(self):
        results = []

        # Setup the script group
        result = self.run_setup()
        if result:
            results.append(result)
            yield result
            if (results[-1].exceptions and Group.abort_on_exception) or results[-1].stopped:
                raise StopScript

        # Run all the scripts
        for method_name in self.__class__.scripts():
            results.append(self.run_script(method_name))
            yield results[-1]
            if (results[-1].exceptions and Group.abort_on_exception) or results[-1].stopped:
                raise StopScript

        # Teardown the script group
        result = self.run_teardown()
        if result:
            results.append(result)
            yield result
            if (results[-1].exceptions and Group.abort_on_exception) or results[-1].stopped:
                raise StopScript
        return results

    # Run a specific script method
    def run_script(self, method_name):
        ScriptStatus.instance().status = f"{self.__class__.__name__} : {method_name}"
        return self.run_method(self, method_name)

    def run_method(self, object, method_name):
        result = ScriptResult()
        Group.current_result = result

        # Verify script method exists
        if hasattr(object.__class__, method_name):
            if hasattr(self, "output_io"):
                self.output_io = self.output_io or StringIO("")
            else:
                self.output_io = StringIO("")

            # Capture STDOUT and STDERR
            sys.stdout = Stdout.instance()
            sys.stderr = Stderr.instance()
            sys.stdout.add_stream(self.output_io)
            sys.stderr.add_stream(self.output_io)

            result.group = object.__class__.__name__
            result.script = method_name
            try:
                method = getattr(object, method_name)
                method()
                result.result = "PASS"

                if openc3.script.RUNNING_SCRIPT and openc3.script.RUNNING_SCRIPT.instance.exceptions:
                    result.exceptions = openc3.script.RUNNING_SCRIPT.instance.exceptions
                    result.result = "FAIL"
                    openc3.script.RUNNING_SCRIPT.instance.exceptions = None

            except Exception as error:
                if isinstance(error, StopScript):
                    result.stopped = True
                    result.result = "STOP"
                if isinstance(error, SkipScript):
                    result.result = "SKIP"
                    if hasattr(error, "message"):
                        result.message = result.message or ""
                        result.message += error.message + "\n"
                else:
                    if not isinstance(error, StopScript) and (
                        not openc3.script.RUNNING_SCRIPT
                        or not openc3.script.RUNNING_SCRIPT.instance
                        or not openc3.script.RUNNING_SCRIPT.instance.exceptions
                        or error not in openc3.script.RUNNING_SCRIPT.instance.exceptions
                    ):
                        result.exceptions = result.exceptions or []
                        result.exceptions.append(error)
                    if (
                        openc3.script.RUNNING_SCRIPT
                        and openc3.script.RUNNING_SCRIPT.instance
                        and openc3.script.RUNNING_SCRIPT.instance.exceptions
                    ):
                        result.exceptions = result.exceptions or []
                        result.exceptions.extend(openc3.script.RUNNING_SCRIPT.instance.exceptions)
                        openc3.script.RUNNING_SCRIPT.instance.exceptions = None
                if result.exceptions:
                    result.result = "FAIL"
            finally:
                result.output = self.output_io.getvalue()
                self.output_io.truncate(0)
                self.output_io.seek(0)
                sys.stdout.remove_stream(self.output_io)
                sys.stdout.remove_stream(self.output_io)

                match result.result:
                    case "FAIL":
                        ScriptStatus.instance().fail_count += 1
                    case "SKIP":
                        ScriptStatus.instance().skip_count += 1
                    case "PASS":
                        ScriptStatus.instance().pass_count += 1
        else:
            Group.current_result = None
            raise Exception(f"Unknown method {method_name} for {object.__class__}")
        Group.current_result = None
        return result

    def run_setup(self):
        result = None
        if "setup" in dir(self):
            ScriptStatus.instance().status = f"{self.__class__} : setup"
            result = self.run_script("setup")
        return result

    def run_teardown(self):
        result = None
        if "teardown" in dir(self):
            ScriptStatus.instance().status = f"{self.__class__} : teardown"
            result = self.run_script("teardown")
        return result

    @classmethod
    def get_num_scripts(cls):
        num_scripts = 0
        if "setup" in dir(cls):
            num_scripts += 1
        if "teardown" in dir(cls):
            num_scripts += 1
        num_scripts += len(cls.scripts())
        return num_scripts

    @classmethod
    def puts(cls, string):
        print(string, file=sys.stdout)
        if Group.current_result:
            Group.current_result.message = Group.current_result.message or ""
            Group.current_result.message += string.rstrip("\n")
            Group.current_result.message += "\n"

    @classmethod
    def print(cls, string):
        print(string, file=sys.stdout)
        if Group.current_result:
            Group.current_result.message = Group.current_result.message or ""
            Group.current_result.message += string.rstrip("\n")
            Group.current_result.message += "\n"

    @classmethod
    def current_suite(cls):
        if Group.current_result:
            return Group.current_result.suite
        else:
            return None

    @classmethod
    def current_group(cls):
        if Group.current_result:
            return Group.current_result.group
        else:
            return None

    @classmethod
    def current_script(cls):
        if Group.current_result:
            return Group.current_result.script
        else:
            return None


# Helper class to collect information about the running scripts like pass / fail counts
class ScriptStatus:
    instance_obj = None

    def __init__(self):
        self.status = ""
        self.pass_count = 0
        self.skip_count = 0
        self.fail_count = 0
        self.total = 1

    @classmethod
    def instance(cls):
        if ScriptStatus.instance_obj:
            return ScriptStatus.instance_obj
        ScriptStatus.instance_obj = cls()
        return ScriptStatus.instance_obj


# Helper class to collect script result information
class ScriptResult:
    suite = None

    def __init__(self):
        self.suite = None
        if ScriptResult.suite:
            ScriptResult.suite = copy.deepcopy(ScriptResult.suite)
        self.group = None
        self.script = None
        self.output = None
        self.exceptions = None
        self.stopped = False
        self.result = "SKIP"
        self.message = None



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import zlib
import time
import threading
from openc3.utilities.bucket import Bucket
from openc3.utilities.logger import Logger
from openc3.models.reducer_model import ReducerModel
from openc3.environment import OPENC3_LOGS_BUCKET

class BucketUtilities:
    FILE_TIMESTAMP_FORMAT = "%Y%m%d%H%M%S%N"
    DIRECTORY_TIMESTAMP_FORMAT = "%Y%m%d"

    # TODO: This is never called
    # @classmethod
    # def bucket_load(cls, *args, scope=OPENC3_SCOPE):
    #     if not scope:
    #         scope = OPENC3_SCOPE
    #     path = args[0]

    #     # Only support TARGET files
    #     if path[0] == "/" or str(path.split("/")[0]).upper() != path.split("/")[0]:
    #         raise ImportError(f"only relative TARGET files are allowed -- {path}")
    #     extension = os.path.splitext(path)[1]
    #     if not extension or extension == "":
    #         path = path + ".py"

    #     # Retrieve the text of the script from S3
    #     text = TargetFile.body(scope, path)
    #     if not text:
    #         raise ImportError(f"Bucket file {path} not found for scope {scope}")

    #     # Execute the script directly without instrumentation because we are doing require/load
    #     exec(text.decode())

    #     # Successful load/require returns true
    #     return True

    @classmethod
    def move_log_file_to_bucket_thread(cls, filename, bucket_key, metadata={}):
        try:
            client = Bucket.getClient()

            orig_filename = None
            if os.path.splitext(filename)[1] != ".txt":
                orig_filename = filename
                filename = cls.compress_file(filename)
                bucket_key += ".gz"

            retry_count = 0
            while retry_count < 3:
                try:
                    # We want to open this as a file and pass that to put_object to allow
                    # this to work with really large files. Otherwise the entire file has
                    # to be held in memory!
                    with open(filename, "rb") as file:
                        client.put_object(
                            bucket=OPENC3_LOGS_BUCKET,
                            key=bucket_key,
                            body=file,
                            metadata=metadata,
                        )
                    break
                except Exception as err:
                    # Try to upload file three times
                    retry_count += 1
                    if retry_count >= 3:
                        raise err
                    Logger.warn(f"Error saving log file to bucket - retry {retry_count}: {filename}\n{str(err)}")
                    time.sleep(1)

            Logger.debug(f"wrote {OPENC3_LOGS_BUCKET}/{bucket_key}")
            ReducerModel.add_file(bucket_key)  # Record the new file for data reduction

            if orig_filename:
                os.remove(orig_filename)
            os.remove(filename)
        except Exception as err:
            Logger.error(f"Error saving log file to bucket: {filename}\n{str(err)}")

    @classmethod
    def move_log_file_to_bucket(cls, filename, bucket_key, metadata={}):
        thread = threading.Thread(
            target=cls.move_log_file_to_bucket_thread,
            args=[filename, bucket_key, metadata],
            daemon=True,
        )
        thread.start()
        return thread

    @classmethod
    def compress_file(cls, filename, chunk_size=50_000_000):
        zipped = f"{filename}.gz"

        obj = zlib.compressobj()
        with open(zipped, "wb") as zip_file:
            with open(filename, "rb") as file:
                while True:
                    chunk = file.read(chunk_size)
                    if chunk:
                        compressed = obj.compress(chunk)
                        zip_file.write(compressed)
                    else:
                        compressed = obj.flush()
                        if compressed:
                            zip_file.write(compressed)
                        break

        return zipped



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1953 and https://github.com/OpenC3/cosmos/pull/1963

from datetime import datetime, timezone
from openc3.packets.packet import Packet
from openc3.utilities.string import simple_formatted
from openc3.utilities.extract import convert_to_value

class Commands:
    """Commands uses PacketConfig to parse the command and telemetry
    configuration files. It contains all the knowledge of which command packets
    exist in the system and how to access them. This class is the API layer
    which other classes use to access commands.

    This should not be confused with the Api module which implements the JSON
    API that is used by tools when accessing the Server. The Api module always
    provides Ruby primitives where the PacketConfig class can return actual
    Packet or PacketItem objects. While there are some overlapping methods between
    the two, these are separate interfaces into the system."""

    LATEST_PACKET_NAME = "LATEST"

    # @param config [PacketConfig] Packet configuration to use to access the
    #  commands
    def __init__(self, config, system):
        self.config = config
        self.system = system

    # (see PacketConfig#warnings)
    def warnings(self):
        return self.config.warnings

    # @return [Array<String>] The command target names (excluding UNKNOWN)
    def target_names(self):
        result = list(self.config.commands.keys())
        if "UNKNOWN" in result:
            result.remove("UNKNOWN")
        return result

    # @param target_name [String] The target name
    # @return [Hash<packet_name=>Packet>] Hash of the command packets for the given
    #   target name keyed by the packet name
    def packets(self, target_name):
        target_packets = self.config.commands.get(target_name.upper(), None)
        if target_packets is None:
            raise RuntimeError(f"Command target '{target_name.upper()}' does not exist")
        return target_packets

    # @param target_name [String] The target name
    # @param packet_name [String] The packet name. Must be a defined packet name
    #   and not 'LATEST'.
    # @return [Packet] The command packet for the given target and packet name
    def packet(self, target_name, packet_name):
        target_packets = self.packets(target_name)
        packet = target_packets.get(packet_name.upper(), None)
        if packet is None:
            raise RuntimeError(f"Command packet '{target_name.upper()} {packet_name.upper()}' does not exist")
        return packet

    # @param target_name (see #packet)
    # @param packet_name (see #packet)
    # @return [Array<PacketItem>] The command parameters for the given target and packet name
    def params(self, target_name, packet_name):
        return self.packet(target_name, packet_name).sorted_items

    # Identifies an unknown buffer of data as a defined command and sets the
    # commands's data to the given buffer. Identifying a command uses the fields
    # marked as ID_PARAMETER to identify if the buffer passed represents the
    # command defined. Incorrectly sized buffers are still processed but an
    # error is logged.
    #
    # Note: Subsequent requests for the command (using packet) will return
    # an uninitialized copy of the command. Thus you must use the return value
    # of this method.
    #
    # Note: this method does not increment received_count and it should be
    # incremented externally if needed.
    #
    # @param (see #identify_tlm!)
    # @return (see #identify_tlm!)
    def identify(self, packet_data, target_names=None):
        identified_packet = None

        if not target_names:
            target_names = self.target_names()

        for target_name in target_names:
            target_name = str(target_name).upper()
            target_packets = None
            try:
                target_packets = self.packets(target_name)
            except RuntimeError:
                # No commands for this target
                continue

            target = self.system.targets.get(target_name)
            if target and target.cmd_unique_id_mode:
                # Iterate through the packets and see if any represent the buffer
                for _, packet in target_packets.items():
                    if packet.identify(packet_data):
                        identified_packet = packet
                        break
            else:
                # Do a lookup to quickly identify the packet
                if len(target_packets) > 0:
                    packet = next(iter(target_packets.values()))
                    key = packet.read_id_values(packet_data)
                    id_values = self.config.cmd_id_value_hash[target_name]
                    identified_packet = id_values.get(str(key))
                    if identified_packet is None:
                        identified_packet = id_values.get("CATCHALL")

            if identified_packet is not None:

                identified_packet = identified_packet.clone()
                identified_packet.received_time = None
                identified_packet.stored = False
                identified_packet.extra = None
                identified_packet.buffer = packet_data
                break

        return identified_packet

    # Returns a copy of the specified command packet with the parameters
    # initialized to the given params values.
    #
    # Note: this method does not increment received_count and it should be
    # incremented externally if needed.
    #
    # @param target_name (see #packet)
    # @param packet_name (see #packet)
    # @param params [Hash<param_name=>param_value>] Parameter items to override
    #   in the given command.
    # @param range_checking [Boolean] Whether to perform range checking on the
    #   passed in parameters.
    # @param raw [Boolean] Indicates whether or not to run conversions on command parameters
    # @param check_required_params [Boolean] Indicates whether or not to check
    #   that the required command parameters are present
    def build_cmd(
        self,
        target_name,
        packet_name,
        params={},
        range_checking=True,
        raw=False,
        check_required_params=True,
    ):
        target_upcase = target_name.upper()
        packet_upcase = packet_name.upper()

        # Lookup the command and create a light weight copy
        pkt = self.packet(target_upcase, packet_upcase)

        command = pkt.clone()

        # Restore the command's buffer to a zeroed string of defined length
        # This will undo any side effects from earlier commands that may have altered the size
        # of the buffer
        command.buffer = bytearray(b"\x00" * command.defined_length)

        # Set time, parameters, and restore defaults
        command.received_time = datetime.now(timezone.utc)
        command.stored = False
        command.extra = None
        command.given_values = params
        command.restore_defaults(command.buffer_no_copy(), list(params.keys()))
        command.raw = raw

        given_item_names = self._set_parameters(command, params, range_checking)
        if check_required_params:
            self._check_required_params(command, given_item_names)

        return command

    # Formatted version of a command
    def format(self, packet, ignored_parameters=[]):
        if packet.raw:
            items = packet.read_all("RAW")
            raw = True
        else:
            items = packet.read_all("FORMATTED")
            raw = False
        items = [item for item in items if item[0] not in ignored_parameters]
        return self.build_cmd_output_string(packet.target_name, packet.packet_name, items, raw)

    def build_cmd_output_string(self, target_name, cmd_name, cmd_params, raw=False):
        if raw:
            output_string = 'cmd_raw("'
        else:
            output_string = 'cmd("'
        if not target_name:
            target_name = "UNKNOWN"
        if not cmd_name:
            cmd_name = "UNKNOWN"
        output_string += target_name + " " + cmd_name
        if cmd_params is None or len(cmd_params) == 0:
            output_string += '")'
        else:
            try:
                command_items = self.packet(target_name, cmd_name).items
            except RuntimeError:
                command_items = {}

            params = []
            for key, value in cmd_params:
                if key in Packet.RESERVED_ITEM_NAMES:
                    continue

                try:
                    item_type = command_items[key].data_type
                except KeyError:
                    item_type = None

                if isinstance(value, str):
                    if item_type == "BLOCK" or item_type == "STRING":
                        if not value.isascii():
                            value = "0x" + simple_formatted(value)
                        else:
                            value = f"'{str(value)}'"
                    else:
                        value = str(convert_to_value(value))
                    if len(value) > 256:
                        value = value[:256] + "...'"
                    value = value.replace('"', "'")
                elif isinstance(value, list):
                    value = f"[{', '.join(str(i) for i in value)}]"
                params.append(f"{key} {value}")
            params = (", ").join(params)
            output_string += " with " + params + '")'
        return output_string

    # Returns whether the given command is hazardous. Commands are hazardous
    # if they are marked hazardous overall or if any of their hardardous states
    # are set. Thus any given parameter values are first applied to the command
    # and then checked for hazardous states.
    #
    # @param command [Packet] The command to check for hazardous
    def cmd_pkt_hazardous(self, command):
        if command.hazardous:
            return (True, command.hazardous_description)

        # Check each item for hazardous states
        for item_name, item_def in command.items.items():
            if item_def.hazardous:
                state_name = command.read(item_name)
                # Nominally the command.read will return a valid state_name
                # If it doesn't, the if check will fail and we'll fall through to
                # the bottom where we return [false, nil] which means this
                # command is not hazardous.
                if item_def.hazardous.get(state_name) is not None:
                    return (True, item_def.hazardous[state_name])

        return (False, None)

    # Returns whether the given command is hazardous. Commands are hazardous
    # if they are marked hazardous overall or if any of their hardardous states
    # are set. Thus any given parameter values are first applied to the command
    # and then checked for hazardous states.
    #
    # @param target_name (see #packet)
    # @param packet_name (see #packet)
    # @param params (see #build_cmd)
    def cmd_hazardous(self, target_name, packet_name, params={}):
        # Build a command without range checking, perform conversions, and don't
        # check required parameters since we're not actually using the command.
        return self.cmd_pkt_hazardous(self.build_cmd(target_name, packet_name, params, False, False, False))

    def all(self):
        return self.config.commands

    def dynamic_add_packet(self, packet, affect_ids=False):
        self.config.dynamic_add_packet(packet, "COMMAND", affect_ids=affect_ids)

    def _set_parameters(self, command, params, range_checking):
        given_item_names = []
        for item_name, value in params.items():
            item_upcase = item_name.upper()
            item = command.get_item(item_upcase)
            range_check_value = value

            if range_checking:
                if item.states:
                    if item.states.get(str(value).upper()) is not None:
                        range_check_value = item.states[str(value).upper()]
                    else:
                        if value not in item.states.values():
                            if command.raw:
                                # Raw commands report missing value maps
                                raise RuntimeError(
                                    f"Command parameter '{command.target_name} {command.packet_name} {item_upcase}' = {value} not one of {', '.join(map(str, item.states.values()))}"
                                )
                            else:
                                # Normal commands report missing state maps
                                raise RuntimeError(
                                    f"Command parameter '{command.target_name} {command.packet_name} {item_upcase}' = {value} not one of {', '.join(item.states.keys())}"
                                )

                # Only range check if we have a min, max and not a string default value
                minimum = item.minimum
                maximum = item.maximum
                if minimum is not None and maximum is not None and not isinstance(item.default, str):
                    # Perform Range Check on command parameter
                    if isinstance(range_check_value, str) or range_check_value < minimum or range_check_value > maximum:
                        if isinstance(range_check_value, str):
                            range_check_value = f"'{range_check_value}'"
                        raise RuntimeError(
                            f"Command parameter '{command.target_name} {command.packet_name} {item_upcase}' = {range_check_value} not in valid range of {minimum} to {maximum}"
                        )

            # Update parameter in command
            if command.raw:
                command.write(item_upcase, value, "RAW")
            else:
                command.write(item_upcase, value, "CONVERTED")
            given_item_names.append(item_upcase)

        return given_item_names

    def _check_required_params(self, command, given_item_names):
        # Script Runner could call this command with only some parameters
        # so make sure any required parameters were actually passed in.
        for item_name, item_def in command.items.items():
            if item_def.required and item_name not in given_item_names:
                raise RuntimeError(
                    f"Required command parameter '{command.target_name} {command.packet_name} {item_name}' not given"
                )



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


class PacketItemLimits:
    """Maintains knowledge of limits for a PacketItem"""

    # Array of all limit states
    LIMITS_STATES = [
        "RED",
        "RED_HIGH",
        "RED_LOW",
        "YELLOW",
        "YELLOW_HIGH",
        "YELLOW_LOW",
        "GREEN",
        "GREEN_HIGH",
        "GREEN_LOW",
        "BLUE",
        "STALE",
        None,
    ]
    # Array of all limit states which should be considered in error
    OUT_OF_LIMITS_STATES = [
        "RED",
        "RED_HIGH",
        "RED_LOW",
        "YELLOW",
        "YELLOW_HIGH",
        "YELLOW_LOW",
    ]

    def __init__(self):
        self.values = None
        self.enabled = False
        self.state = None
        self.response = None
        self.persistence_setting = 1
        self.persistence_count = 0

    @property
    def values(self):
        return self.__values

    @values.setter
    def values(self, values):
        if values is not None:
            if not isinstance(values, dict):
                raise TypeError(f"values must be a Hash but is a {values.__class__.__name__}")
            if "DEFAULT" not in values:
                raise ValueError("values must be a Hash with a 'DEFAULT' key")
            self.__values = values
        else:
            self.__values = None

    @property
    def state(self):
        return self.__state

    @state.setter
    def state(self, state):
        if state not in PacketItemLimits.LIMITS_STATES:
            raise ValueError(f"state must be one of {PacketItemLimits.LIMITS_STATES} but is {state}")
        self.__state = state

    @property
    def response(self):
        return self.__response

    @response.setter
    def response(self, response):
        if response is not None:
            if "LimitsResponse" not in response.__class__.__name__:
                raise TypeError(f"response must be a LimitsResponse but is a {response.__class__.__name__}")
            self.__response = response
        else:
            self.__response = None

    def as_json(self):
        limits = {}
        limits['values'] = self.__values
        limits['enabled'] = self.enabled
        limits['state'] = self.__state
        if self.__response:
            limits['response'] = str(self.__response)
        else:
            limits['response'] = None
        limits['persistence_setting'] = self.persistence_setting
        limits['persistence_count'] = self.persistence_count
        return limits



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from .stream_log import StreamLog


class StreamLogPair:
    """Holds a read/write pair of stream logs"""

    # self.param name [String] name to be added to log filenames
    # self.param params [Array] stream log writer parameters or empty array
    def __init__(self, name, params=[]):
        self.read_log = StreamLog(name, "READ", *params)
        self.write_log = StreamLog(name, "WRITE", *params)

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = name
        self.read_log.name = name
        self.write_log.name = name

    # Start stream logs
    def start(self):
        self.read_log.start()
        self.write_log.start()

    # Close any open stream log files
    def stop(self):
        self.read_log.stop()
        self.write_log.stop()

    def shutdown(self):
        self.read_log.shutdown()
        self.write_log.shutdown()

    # TODO: Simply copy.copy
    # Clone the stream log pair
    # def clone(self):
    #   stream_log_pair = super.clone()
    #   stream_log_pair.read_log = self.read_log.clone
    #   stream_log_pair.write_log = self.write_log.clone
    #   stream_log_pair.read_log.start if self.read_log.logging_enabled:
    #   stream_log_pair.write_log.start if self.write_log.logging_enabled:
    #   stream_log_pair



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
import time
from typing import Any

from openc3.utilities.store import Store
from openc3.utilities.store_queued import StoreQueued
from openc3.models.model import Model
from openc3.models.target_model import TargetModel
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.json import JsonEncoder, JsonDecoder


class CvtModel(Model):
    packet_cache = {}
    override_cache = {}

    VALUE_TYPES = {"RAW", "CONVERTED", "FORMATTED", "WITH_UNITS"}

    @classmethod
    def build_json_from_packet(cls, packet):
        return packet.decom()

    @classmethod
    def delete(cls, target_name: str, packet_name: str, scope: str = OPENC3_SCOPE):
        """Delete the current value table for a target"""
        key = f"{scope}__tlm__{target_name}"
        tgt_pkt_key = key + f"__{packet_name}"
        CvtModel.packet_cache[tgt_pkt_key] = None
        Store.hdel(key, packet_name)

    @classmethod
    def set(cls, hash: dict, target_name: str, packet_name: str, queued: bool = False, scope: str = OPENC3_SCOPE):
        """Set the current value table for a target, packet"""
        packet_json = json.dumps(hash, cls=JsonEncoder)
        key = f"{scope}__tlm__{target_name}"
        tgt_pkt_key = key + f"__{packet_name}"
        CvtModel.packet_cache[tgt_pkt_key] = [time.time(), hash]
        if queued:
            StoreQueued.hset(key, packet_name, packet_json)
        else:
            Store.hset(key, packet_name, packet_json)

    # Get the dict for packet in the CVT
    # Note: Does not apply overrides
    @classmethod
    def get(cls, target_name: str, packet_name: str, cache_timeout: float = 0.1, scope: str = OPENC3_SCOPE):
        key = f"{scope}__tlm__{target_name}"
        tgt_pkt_key = key + f"__{packet_name}"
        now = time.time()
        if tgt_pkt_key in CvtModel.packet_cache:
            cache_time, pkt_hash = CvtModel.packet_cache[tgt_pkt_key]
            if (now - cache_time) < cache_timeout:
                return pkt_hash
        packet = Store.hget(key, packet_name)
        if packet is None:
            raise RuntimeError(f"Packet '{target_name} {packet_name}' does not exist")
        pkt_hash = json.loads(packet, cls=JsonDecoder)
        CvtModel.packet_cache[tgt_pkt_key] = [now, pkt_hash]
        return pkt_hash

    # Set an item in the current value table
    @classmethod
    def set_item(
        cls,
        target_name: str,
        packet_name: str,
        item_name: str,
        value: Any,
        type: str,
        queued: bool = False,
        scope: str = OPENC3_SCOPE,
    ):
        pkt_hash = cls.get(target_name, packet_name, cache_timeout=0.0, scope=scope)
        match type:
            case "WITH_UNITS":
                pkt_hash[f"{item_name}__U"] = str(value)  # WITH_UNITS should always be a string
            case "FORMATTED":
                pkt_hash[f"{item_name}__F"] = str(value)  # FORMATTED should always be a string
            case "CONVERTED":
                pkt_hash[f"{item_name}__C"] = value
            case "RAW":
                pkt_hash[item_name] = value
            case "ALL":
                pkt_hash[f"{item_name}__U"] = str(value)  # WITH_UNITS should always be a string
                pkt_hash[f"{item_name}__F"] = str(value)  # FORMATTED should always be a string
                pkt_hash[f"{item_name}__C"] = value
                pkt_hash[item_name] = value
            case _:
                raise RuntimeError(f"Unknown type '{type}' for {target_name} {packet_name} {item_name}")
        cls.set(pkt_hash, target_name=target_name, packet_name=packet_name, queued=queued, scope=scope)

    # Get an item from the current value table
    @classmethod
    def get_item(
        cls,
        target_name,
        packet_name,
        item_name,
        type,
        cache_timeout=0.1,
        scope=OPENC3_SCOPE,
    ):
        result, types = cls._handle_item_override(
            target_name,
            packet_name,
            item_name,
            type=type,
            cache_timeout=cache_timeout,
            scope=scope,
        )
        if result is not None:
            return result
        pkt_hash = cls.get(target_name, packet_name, scope=scope)
        for cvt_value in [pkt_hash[x] for x in types if x in pkt_hash]:
            if cvt_value is not None:
                if type == "FORMATTED" or type == "WITH_UNITS":
                    return str(cvt_value)
                return cvt_value
        # RECEIVED_COUNT is a special case where it is 0 if it doesn't exist
        # This allows scripts to check against the value to see if the packet was ever received
        if item_name == "RECEIVED_COUNT":
            return 0
        else:
            return None

    # Return all item values and limit state from the CVT
    #
    # @param items [Array<String>] Items to return. Must be formatted as TGT__PKT__ITEM__TYPE
    # @param stale_time [Integer] Time in seconds from Time.now that value will be marked stale
    # @return [Array] Array of values
    @classmethod
    def get_tlm_values(cls, items: list, stale_time: int = 30, cache_timeout: float = 0.1, scope: str = OPENC3_SCOPE):
        now = time.time()
        results = []
        lookups = []
        packet_lookup = {}
        overrides = {}
        # First generate a lookup dict of all the items represented so we can query the CVT
        for item in items:
            cls._parse_item(now, lookups, overrides, item, cache_timeout=cache_timeout, scope=scope)

        for target_packet_key, target_name, packet_name, value_keys in lookups:
            if target_packet_key not in packet_lookup:
                packet_lookup[target_packet_key] = cls.get(
                    target_name,
                    packet_name,
                    cache_timeout,
                    scope,
                )
            pkt_hash = packet_lookup[target_packet_key]
            item_result = []
            if isinstance(value_keys, dict):  # Set in _parse_item to indicate override
                item_result.insert(0, value_keys["value"])
            else:
                for key in value_keys:
                    if key in pkt_hash:
                        item_result.insert(0, pkt_hash[key])
                        break  # We want the first value
                # If we were able to find a value, try to get the limits state
                if len(item_result) > 0 and item_result[0] is not None:
                    if now - pkt_hash["RECEIVED_TIMESECONDS"] > stale_time:
                        item_result.insert(1, "STALE")
                    else:
                        # The last key is simply the name (RAW) so we can append __L
                        # If there is no limits then it returns None which is acceptable
                        item_result.insert(1, pkt_hash.get(f"{value_keys[-1]}__L"))
                else:
                    if value_keys[-1] not in pkt_hash:
                        raise RuntimeError(f"Item '{target_name} {packet_name} {value_keys[-1]}' does not exist")
                    else:
                        item_result.insert(1, None)
            results.append(item_result)
        return results

    @classmethod
    def overrides(cls, scope=OPENC3_SCOPE):
        """Return all the overrides"""
        overrides = []
        for target_name in TargetModel.names(scope):
            all = Store.hgetall(f"{scope}__override__{target_name}")
            if len(all) == 0:
                continue
            # decode the binary string keys to strings
            all = {k.decode(): v for (k, v) in all.items()}
            for packet_name, pkt_hash in all.items():
                items = json.loads(pkt_hash, cls=JsonDecoder)
                for key, value in items.items():
                    item = {}
                    item["target_name"] = target_name
                    item["packet_name"] = packet_name
                    if "__" in key:
                        item_name, value_type_key = key.split("__")
                    else:  # RAW item which doesn't have an underscore
                        item_name = key
                        value_type_key = "R"
                    item["item_name"] = item_name
                    match value_type_key:
                        case "U":
                            item["value_type"] = "WITH_UNITS"
                        case "F":
                            item["value_type"] = "FORMATTED"
                        case "C":
                            item["value_type"] = "CONVERTED"
                        case "R":
                            item["value_type"] = "RAW"
                    item["value"] = value
                    overrides.append(item)
        return overrides

    @classmethod
    def override(cls, target_name, packet_name, item_name, value, type="ALL", scope=OPENC3_SCOPE):
        """Override a current value table item such that it always returns the same value for the given type"""
        pkt_hash = Store.hget(f"{scope}__override__{target_name}", packet_name)
        if pkt_hash is not None:
            pkt_hash = json.loads(pkt_hash)
        else:
            pkt_hash = {}
        match type:
            case "ALL":
                pkt_hash[item_name] = value
                pkt_hash[f"{item_name}__C"] = value
                pkt_hash[f"{item_name}__F"] = str(value)
                pkt_hash[f"{item_name}__U"] = str(value)
            case "RAW":
                pkt_hash[item_name] = value
            case "CONVERTED":
                pkt_hash[f"{item_name}__C"] = value
            case "FORMATTED":
                pkt_hash[f"{item_name}__F"] = str(value)  # Always a String
            case "WITH_UNITS":
                pkt_hash[f"{item_name}__U"] = str(value)  # Always a String
            case _:
                raise RuntimeError(f"Unknown type '{type}' for {target_name} {packet_name} {item_name}")
        tgt_pkt_key = f"{scope}__tlm__{target_name}__{packet_name}"
        CvtModel.override_cache[tgt_pkt_key] = [time.time(), pkt_hash]
        Store.hset(f"{scope}__override__{target_name}", packet_name, json.dumps(pkt_hash))

    # Normalize a current value table item such that it returns the actual value
    @classmethod
    def normalize(cls, target_name, packet_name, item_name, type="ALL", scope=OPENC3_SCOPE):
        pkt_hash = Store.hget(f"{scope}__override__{target_name}", packet_name)
        if pkt_hash is not None:
            pkt_hash = json.loads(pkt_hash)
        else:
            pkt_hash = {}
        match type:
            case "ALL":
                pkt_hash.pop(item_name, None)
                pkt_hash.pop(f"{item_name}__C", None)
                pkt_hash.pop(f"{item_name}__F", None)
                pkt_hash.pop(f"{item_name}__U", None)
            case "RAW":
                if item_name in pkt_hash:
                    pkt_hash.pop(item_name)
            case "CONVERTED":
                if f"{item_name}__C" in pkt_hash:
                    pkt_hash.pop(f"{item_name}__C")
            case "FORMATTED":
                if f"{item_name}__F" in pkt_hash:
                    pkt_hash.pop(f"{item_name}__F")
            case "WITH_UNITS":
                if f"{item_name}__U" in pkt_hash:
                    pkt_hash.pop(f"{item_name}__U")
            case _:
                raise RuntimeError(f"Unknown type '{type}' for {target_name} {packet_name} {item_name}")
        tgt_pkt_key = f"{scope}__tlm__{target_name}__{packet_name}"
        if len(pkt_hash) == 0:
            if tgt_pkt_key in CvtModel.override_cache:
                CvtModel.override_cache.pop(tgt_pkt_key)
            Store.hdel(f"{scope}__override__{target_name}", packet_name)
        else:
            CvtModel.override_cache[tgt_pkt_key] = [time.time(), pkt_hash]
            Store.hset(f"{scope}__override__{target_name}", packet_name, json.dumps(pkt_hash))

    @classmethod
    def determine_latest_packet_for_item(cls, target_name, item_name, cache_timeout=0.1, scope=OPENC3_SCOPE):
        item_map = TargetModel.get_item_to_packet_map(target_name, scope=scope)
        packet_names = item_map.get(item_name)
        if packet_names is None:
            raise RuntimeError(f"Item '{target_name} LATEST {item_name}' does not exist for scope: {scope}")

        latest = -1
        latest_packet_name = None
        for packet_name in packet_names:
            pkt_hash = cls.get(
                target_name,
                packet_name,
                cache_timeout,
                scope,
            )
            if pkt_hash["PACKET_TIMESECONDS"] and pkt_hash["PACKET_TIMESECONDS"] > latest:
                latest = pkt_hash["PACKET_TIMESECONDS"]
                latest_packet_name = packet_name
        if latest == -1:
            raise RuntimeError(f"Item '{target_name} LATEST {item_name}' does not exist for scope: {scope}")
        return latest_packet_name

    @classmethod
    def _handle_item_override(
        cls,
        target_name,
        packet_name,
        item_name,
        type,
        cache_timeout,
        scope=OPENC3_SCOPE,
    ):
        override_key = item_name
        types = []
        match type:
            case "WITH_UNITS":
                types = [
                    f"{item_name}__U",
                    f"{item_name}__F",
                    f"{item_name}__C",
                    item_name,
                ]
                override_key = f"{item_name}__U"
            case "FORMATTED":
                types = [f"{item_name}__F", f"{item_name}__C", item_name]
                override_key = f"{item_name}__F"
            case "CONVERTED":
                types = [f"{item_name}__C", item_name]
                override_key = f"{item_name}__C"
            case "RAW":
                types = [item_name]
            case _:
                raise RuntimeError(f"Unknown type '{type}' for {target_name} {packet_name} {item_name}")

        tgt_pkt_key = f"{scope}__tlm__{target_name}__{packet_name}"
        overrides = cls._get_overrides(
            time.time(),
            tgt_pkt_key,
            {},
            target_name,
            packet_name,
            cache_timeout=cache_timeout,
            scope=scope,
        )
        result = overrides.get(override_key)
        if result is not None:
            return result, types
        return None, types

    @classmethod
    def _get_overrides(cls, now, tgt_pkt_key, overrides, target_name, packet_name, cache_timeout, scope):
        if tgt_pkt_key in CvtModel.override_cache:
            cache_time, pkt_hash = CvtModel.override_cache[tgt_pkt_key]
            if (now - cache_time) < cache_timeout:
                overrides[tgt_pkt_key] = pkt_hash
                return pkt_hash
        override_data = Store.hget(f"{scope}__override__{target_name}", packet_name)
        if override_data is not None:
            pkt_hash = json.loads(override_data)
            overrides[tgt_pkt_key] = pkt_hash
        else:
            pkt_hash = {}
            overrides[tgt_pkt_key] = {}
        CvtModel.override_cache[tgt_pkt_key] = [now, pkt_hash]  # always update
        return pkt_hash

    # parse item and update lookups with packet_name and target_name and keys
    # return an ordered array of dict with keys
    @classmethod
    def _parse_item(cls, now, lookups, overrides, item, cache_timeout, scope):
        target_name, packet_name, item_name, value_type = item

        # We build lookup keys by including all the less formatted types to gracefully degrade lookups
        # This allows the user to specify WITH_UNITS and if there is no conversions it will simply return the RAW value
        match str(value_type):
            case "RAW":
                keys = [item_name]
            case "CONVERTED":
                keys = [f"{item_name}__C", item_name]
            case "FORMATTED":
                keys = [f"{item_name}__F", f"{item_name}__C", item_name]
            case "WITH_UNITS":
                keys = [
                    f"{item_name}__U",
                    f"{item_name}__F",
                    f"{item_name}__C",
                    item_name,
                ]
            case _:
                raise ValueError(f"Unknown value type '{value_type}'")

        # Check the overrides cache for this target / packet
        tgt_pkt_key = f"{scope}__tlm__{target_name}__{packet_name}"
        if tgt_pkt_key not in overrides:
            cls._get_overrides(
                now,
                tgt_pkt_key,
                overrides,
                target_name,
                packet_name,
                cache_timeout=cache_timeout,
                scope=scope,
            )

        # Set the result as a Hash to distinguish it from the key array and from an overridden Array value
        if tgt_pkt_key in overrides and keys[0] in overrides[tgt_pkt_key]:
            keys = {"value": overrides[tgt_pkt_key][keys[0]]}

        lookups.append([tgt_pkt_key, target_name, packet_name, keys])



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
import openc3.script
from openc3.environment import OPENC3_SCOPE

def package_list(scope: str = OPENC3_SCOPE):
    try:
        endpoint = f"/openc3-api/packages?scope={scope}"
        response = openc3.script.API_SERVER.request("get", endpoint, scope=scope)
        if not response or response.status_code != 200:
            raise RuntimeError(f"Unexpected response to package_list: {response}")
        return json.loads(response.text)
    except Exception as error:
        raise RuntimeError(f"package_list failed due to {repr(error)}") from error



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.accessors.accessor import Accessor
import re


class TemplateAccessor(Accessor):
    def __init__(self, packet, left_char="<", right_char=">"):
        super().__init__(packet)
        self.left_char = left_char
        self.right_char = right_char
        self.configured = False

    def configure(self):
        if self.configured:
            return

        escaped_left_char = self.left_char
        if self.left_char == "(":
            escaped_left_char = f"\\{self.left_char}"
        escaped_right_char = self.right_char
        if self.right_char == ")":
            escaped_right_char = f"\\{self.right_char}"

        # Convert the template into a Regexp for reading each item
        template = self.packet.template[:].decode()
        template_items = re.compile(f"{escaped_left_char}.*?{escaped_right_char}", re.X).findall(template)
        escaped_read_template = re.escape(template)

        self.item_keys = []
        for item in template_items:
            self.item_keys.append(item[1:-1])
            # If they're using parens we have to escape them
            # since we're working with the already escaped template
            if self.left_char == "(":
                item = f"\\({item[1:]}"
            if self.right_char == ")":
                item = f"{item[0:-1]}\\)"
            escaped_read_template = escaped_read_template.replace(item, "(.*)")
        self.read_regexp = re.compile(escaped_read_template, re.X)
        self.configured = True

    def read_item(self, item, buffer):
        if item.data_type == "DERIVED":
            return None
        self.configure()

        # Scan the response for all the variables in brackets <VARIABLE>
        values = self.read_regexp.match(buffer.decode())
        if values is not None:
            values = values.groups()
        if values is None or (len(values) != len(self.item_keys)):
            num_items = 0
            if values is not None:
                num_items = len(values)
            raise RuntimeError(
                f"Unexpected number of items found in buffer: {num_items}, Expected: {len(self.item_keys)}"
            )
        else:
            for i, value in enumerate(values):
                item_key = self.item_keys[i]
                if item_key == item.key:
                    return Accessor.convert_to_type(value, item)

        raise RuntimeError(f"Response does not include key {item.key}: {buffer}")

    def read_items(self, items, buffer):
        result = {}
        self.configure()

        # Scan the response for all the variables in brackets <VARIABLE>
        values = self.read_regexp.match(buffer.decode())
        if values is not None:
            values = values.groups()
        if values is None or (len(values) != len(self.item_keys)):
            num_items = 0
            if values is not None:
                num_items = len(values)
            raise RuntimeError(
                f"Unexpected number of items found in buffer: {num_items}, Expected: {len(self.item_keys)}"
            )
        else:
            for item in items:
                if item.data_type == "DERIVED":
                    result[item.name] = None
                    continue
                try:
                    index = self.item_keys.index(item.key)
                    result[item.name] = Accessor.convert_to_type(values[index], item)
                except ValueError:
                    raise RuntimeError(f"Unknown item with key {item.key} requested")

        return result

    def write_item(self, item, value, buffer):
        if item.data_type == "DERIVED":
            return None
        self.configure()

        updated_buffer = buffer.decode().replace(f"{self.left_char}{item.key}{self.right_char}", str(value)).encode()

        if buffer == updated_buffer:
            raise RuntimeError(f"Key {item.key} not found in template")
        buffer[0:] = updated_buffer
        return value

    def write_items(self, items, values, buffer):
        self.configure()
        for index, item in enumerate(items):
            if item.data_type == "DERIVED":
                continue
            updated_buffer = (
                buffer.decode().replace(f"{self.left_char}{item.key}{self.right_char}", str(values[index])).encode()
            )

            if buffer == updated_buffer:
                raise RuntimeError(f"Key {item.key} not found in template")
            buffer[0:] = updated_buffer
        return values

    # If this is set it will enforce that buffer data is encoded
    # in a specific encoding
    def enforce_encoding(self):
        return None

    # This affects whether the Packet class enforces the buffer
    # length at all.  Set to false to remove any correlation between
    # buffer length and defined sizes of items in COSMOS
    def enforce_length(self):
        return False

    # This sets the short_buffer_allowed flag in the Packet class
    # which allows packets that have a buffer shorter than the defined size.
    # Note that the buffer is still resized to the defined length
    def enforce_short_buffer_allowed(self):
        return True

    # If this is true it will enforce that COSMOS DERIVED items must have a
    # write_conversion to be written
    def enforce_derived_write_conversion(self, _item):
        return True



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import tempfile
import threading
from datetime import datetime, timezone, timedelta
from openc3.config.config_parser import ConfigParser
from openc3.top_level import kill_thread
from openc3.topics.topic import Topic
from openc3.utilities.bucket_utilities import BucketUtilities
from openc3.utilities.logger import Logger
from openc3.utilities.sleeper import Sleeper
from openc3.utilities.string import build_timestamped_filename
from openc3.utilities.time import from_nsec_from_epoch, to_timestamp


# Creates a log. Can automatically cycle the log based on an elapsed
# time period or when the log file reaches a predefined size.
class LogWriter:
    # The cycle time interval. Cycle times are only checked at this level of
    # granularity.
    CYCLE_TIME_INTERVAL = 10
    # Delay in seconds before trimming Redis streams
    CLEANUP_DELAY = 60

    # Mutex protecting class variables
    mutex = threading.Lock()

    # Array of instances used to keep track of cycling logs
    instances = []

    # Thread used to cycle logs across all log writers
    cycle_thread = None

    # Sleeper used to delay cycle thread
    cycle_sleeper = None

    # self.param remote_log_directory [String] The path to store the log files
    # self.param logging_enabled [Boolean] Whether to start with logging enabled
    # self.param cycle_time [Integer] The amount of time in seconds before creating
    #   a new log file. This can be combined with cycle_size.
    # self.param cycle_size [Integer] The size in bytes before creating a new log
    #   file. This can be combined with cycle_time.
    # self.param cycle_hour [Integer] The time at which to cycle the log. Combined with
    #   cycle_minute to cycle the log daily at the specified time. If None, the log
    #   will be cycled hourly at the specified cycle_minute.
    # self.param cycle_minute [Integer] The time at which to cycle the log. See cycle_hour
    #   for more information.
    def __init__(
        self,
        remote_log_directory,
        logging_enabled=True,
        cycle_time=None,
        cycle_size=1_000_000_000,
        cycle_hour=None,
        cycle_minute=None,
        enforce_time_order=True,
    ):
        self.remote_log_directory = remote_log_directory
        self.logging_enabled = ConfigParser.handle_true_false(logging_enabled)
        self.cycle_time = ConfigParser.handle_none(cycle_time)
        if self.cycle_time:
            self.cycle_time = int(self.cycle_time)
            if self.cycle_time < LogWriter.CYCLE_TIME_INTERVAL:
                raise RuntimeError(f"cycle_time must be >= {LogWriter.CYCLE_TIME_INTERVAL}")
        self.cycle_size = ConfigParser.handle_none(cycle_size)
        if self.cycle_size:
            self.cycle_size = int(self.cycle_size)
        self.cycle_hour = ConfigParser.handle_none(cycle_hour)
        if self.cycle_hour:
            self.cycle_hour = int(self.cycle_hour)
        self.cycle_minute = ConfigParser.handle_none(cycle_minute)
        if self.cycle_minute:
            self.cycle_minute = int(self.cycle_minute)
        self.enforce_time_order = ConfigParser.handle_true_false(enforce_time_order)
        self.out_of_order = False
        self.mutex = threading.Lock()
        self.file = None
        self.label = None
        self.file_size = 0
        self.filename = None
        self.start_time = datetime.now(timezone.utc)
        self.first_time = None
        self.last_time = None
        self.cancel_threads = False
        self.last_offsets = {}
        self.cleanup_offsets = []
        self.cleanup_times = []
        self.previous_time_nsec_since_epoch = None
        self.tmp_dir = tempfile.TemporaryDirectory()

        # This is an optimization to avoid creating a new entry object
        # each time we create an entry which we do a LOT!
        self.entry = ""

        # Always make sure there is a cycle thread - (because it does trimming)
        with LogWriter.mutex:
            LogWriter.instances.append(self)

            if not LogWriter.cycle_thread:
                LogWriter.cycle_thread = threading.Thread(target=self.cycle_thread_body, daemon=True)
                LogWriter.cycle_thread.start()

    # Starts a new log file by closing the existing log file. New log files are
    # not created until packets are written by {#write} so this does not
    # immediately create a log file on the filesystem.
    def start(self):
        with self.mutex:
            self.close_file(False)
            self.logging_enabled = True

    # Stops all logging and closes the current log file.
    def stop(self):
        threads = None
        with self.mutex:
            threads = self.close_file(False)
            self.logging_enabled = False
        return threads

    # Stop all logging, close the current log file, and kill the logging threads.
    def shutdown(self):
        threads = self.stop()
        with LogWriter.mutex:
            LogWriter.instances.remove(self)
            if len(LogWriter.instances) <= 0:
                if LogWriter.cycle_sleeper:
                    LogWriter.cycle_sleeper.cancel()
                if LogWriter.cycle_thread:
                    kill_thread(self, LogWriter.cycle_thread)
                LogWriter.cycle_thread = None
        # Wait for BucketUtilities to finish move_log_file_to_bucket_thread
        for thread in threads:
            thread.join()
        self.tmp_dir.cleanup()

    def graceful_kill(self):
        self.cancel_threads = True

    # implementation details

    def create_unique_filename(self, ext=".log"):
        # Create a filename that doesn't exist
        attempt = None
        while True:
            filename_parts = [attempt]
            if self.label:
                filename_parts.insert(self.label, 0)
            filename = os.path.join(
                self.tmp_dir.name,
                build_timestamped_filename([self.label, attempt], ext),
            )
            if os.path.exists(filename):
                if attempt is None:
                    attempt = 0
                attempt += 1
                Logger.warn(f"Unexpected file name conflict {filename}")
            else:
                return filename

    def cycle_thread_body(self):
        LogWriter.cycle_sleeper = Sleeper()
        while True:
            start_time = datetime.now(timezone.utc)
            with LogWriter.mutex:
                for instance in LogWriter.instances:
                    # The check against start_time needs to be mutex protected to prevent a packet coming in between the check
                    # and closing the file
                    with instance.mutex:
                        utc_now = datetime.now(timezone.utc)
                        if instance.logging_enabled and instance.filename:  # Logging and file opened
                            # Cycle based on total time logging
                            if (
                                instance.cycle_time
                                and (utc_now - instance.start_time).total_seconds() > instance.cycle_time
                            ):
                                Logger.debug("Log writer start new file due to cycle time")
                                instance.close_file(False)
                            # Cycle daily at a specific time
                            elif (
                                instance.cycle_hour
                                and instance.cycle_minute
                                and utc_now.hour == instance.cycle_hour
                                and utc_now.min == instance.cycle_minute
                                and instance.start_time.day != utc_now.day
                            ):
                                Logger.debug("Log writer start new file daily")
                                instance.close_file(False)
                            # Cycle hourly at a specific time
                            elif (
                                instance.cycle_minute
                                and not instance.cycle_hour
                                and utc_now.min == instance.cycle_minute
                                and instance.start_time.hour != utc_now.hour
                            ):
                                Logger.debug("Log writer start new file hourly")
                                instance.close_file(False)

                        # Check for cleanup time
                        indexes_to_clear = []
                        for index, cleanup_time in enumerate(instance.cleanup_times):
                            if cleanup_time <= utc_now:
                                # Now that the file is in S3, trim the Redis stream up until the previous file.
                                # This keeps one minute of data in Redis
                                for (
                                    redis_topic,
                                    cleanup_offset,
                                ) in instance.cleanup_offsets[index]:
                                    Topic.trim_topic(redis_topic, cleanup_offset)
                                indexes_to_clear.append(index)
                        if len(indexes_to_clear) > 0:
                            for index in indexes_to_clear:
                                instance.cleanup_offsets[index] = None
                                instance.cleanup_times[index] = None
                            instance.cleanup_offsets = [x for x in instance.cleanup_offsets if x is not None]
                            instance.cleanup_times = [x for x in instance.cleanup_times if x is not None]

            # Only check whether to cycle at a set interval
            run_time = (datetime.now(timezone.utc) - start_time).total_seconds()
            sleep_time = LogWriter.CYCLE_TIME_INTERVAL - run_time
            if sleep_time < 0:
                sleep_time = 0
            if self.cancel_threads or LogWriter.cycle_sleeper.sleep(sleep_time):
                break

    # Starting a new log file is a critical operation so the entire method is
    # wrapped with a except: and handled with handle_critical_exception
    # Assumes mutex has already been taken
    def start_new_file(self):
        try:
            if self.file:
                self.close_file(False)

            # Start log file
            self.filename = self.create_unique_filename()
            self.file = open(self.filename, "bx")
            self.file_size = 0

            self.start_time = datetime.now(timezone.utc)
            self.out_of_order = False
            self.first_time = None
            self.last_time = None
            self.previous_time_nsec_since_epoch = None
            Logger.debug(f"Log File Opened : {self.filename}")
        except IOError as error:
            Logger.error(f"Error starting new log file {repr(error)}")
            self.logging_enabled = False
            # TODO: handle_critical_exception(err)

    def prepare_write(
        self,
        time_nsec_since_epoch,
        data_length,
        redis_topic=None,
        redis_offset=None,
        allow_new_file=True,
    ):
        # This check includes logging_enabled again because it might have changed since we acquired the mutex
        # Ensures new files based on size, and ensures always increasing time order in files
        if self.logging_enabled:
            if not self.file:
                Logger.debug("Log writer start new file because no file opened")
                if allow_new_file:
                    self.start_new_file()
            elif self.cycle_size and ((self.file_size + data_length) > self.cycle_size):
                Logger.debug(f"Log writer start new file due to cycle size {self.cycle_size}")
                if allow_new_file:
                    self.start_new_file()
            elif (
                self.enforce_time_order
                and self.previous_time_nsec_since_epoch
                and (self.previous_time_nsec_since_epoch > time_nsec_since_epoch)
            ):
                # Warning= Creating new files here can cause lots of files to be created if packets make it through out of order:
                # Changed to just a error to prevent file thrashing
                if not self.out_of_order:
                    Logger.error(
                        f"Log writer out of order time detected (increase buffer depth?): {from_nsec_from_epoch(self.previous_time_nsec_since_epoch)} {from_nsec_from_epoch(time_nsec_since_epoch)}"
                    )
                    self.out_of_order = True
        # This is needed for the redis offset marker entry at the end of the log file
        if redis_topic and redis_offset:
            self.last_offsets[redis_topic] = redis_offset
        self.previous_time_nsec_since_epoch = time_nsec_since_epoch

    # Closing a log file isn't critical so we just log an error. NOTE: This also trims the Redis stream
    # to keep a full file's worth of data in the stream. This is what prevents continuous stream growth.
    # Returns thread that moves log to bucket
    def close_file(self, take_mutex=True):
        threads = []
        if take_mutex:
            self.mutex.acquire()
        try:
            if self.file:
                self.file.close()
                Logger.debug(f"Log File Closed : {self.filename}")
                date = self.first_timestamp()[0:8]  # YYYYMMDD
                bucket_key = os.path.join(self.remote_log_directory, date, self.bucket_filename())
                # Cleanup timestamps here so they are unset for the next file
                self.first_time = None
                self.last_time = None
                threads.append(BucketUtilities.move_log_file_to_bucket(self.filename, bucket_key))
                # Now that the file is in storage, trim the Redis stream after a delay
                self.cleanup_offsets.append({})
                for redis_topic, last_offset in self.last_offsets:
                    self.cleanup_offsets[-1][redis_topic] = last_offset
                self.cleanup_times.append(datetime.now(timezone.utc) + timedelta(seconds=LogWriter.CLEANUP_DELAY))
                self.last_offsets.clear
                self.file = None
                self.file_size = 0
                self.filename = None
        except RuntimeError as error:
            Logger.error(f"Error closing {self.filename} : {repr(error)}")
        finally:
            if take_mutex:
                self.mutex.release()
        return threads

    def bucket_filename(self):
        return f"{self.first_timestamp()}__{self.last_timestamp()}" + self.extension()

    def extension(self):
        return ".log"

    def first_timestamp(self):
        return to_timestamp(from_nsec_from_epoch(self.first_time))  # "YYYYMMDDHHmmSSNNNNNNNNN"

    def last_timestamp(self):
        return to_timestamp(from_nsec_from_epoch(self.last_time))  # "YYYYMMDDHHmmSSNNNNNNNNN"



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.models.secret_model import SecretModel
from openc3.utilities.secrets import Secrets


class RedisSecrets(Secrets):
    def keys(self, secret_store=None, scope=None):
        return SecretModel.names(scope=scope)

    def get(self, key, secret_store=None, scope=None):
        data = SecretModel.get(name=key, scope=scope)
        if data is not None:
            return data["value"]
        else:
            return None

    def set(self, key, value, secret_store=None, scope=None):
        SecretModel.set({"name": key, "value": str(value)}, scope=scope)

    def delete(self, key, secret_store=None, scope=None):
        model = SecretModel.get_model(name=key, scope=scope)
        if model:
            model.destroy()



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1953 and https://github.com/OpenC3/cosmos/pull/1963

# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1957

import json
import time
from typing import Any
from openc3.environment import OPENC3_SCOPE
from openc3.topics.topic import Topic
from openc3.models.model import Model
from openc3.models.microservice_model import MicroserviceModel
from openc3.utilities.json import JsonEncoder
from openc3.utilities.store import Store
from openc3.utilities.logger import Logger
from openc3.utilities.bucket import Bucket
from openc3.environment import OPENC3_CONFIG_BUCKET


# Manages the target in Redis. It stores the target itself under the
# <SCOPE>__openc3_targets key under the target name field. All the command packets
# in the target are stored under the <SCOPE>__openc3cmd__<TARGET NAME> key and the
# telemetry under the <SCOPE>__openc3tlm__<TARGET NAME> key. Any new limits sets
# are merged into the <SCOPE>__limits_sets key as fields. Any new limits groups are
# created under <SCOPE>__limits_groups with field name. These Redis key/fields are
# all removed when the undeploy method is called.
class TargetModel(Model):
    PRIMARY_KEY = "openc3_targets"
    VALID_TYPES = {"CMD", "TLM"}
    ITEM_MAP_CACHE_TIMEOUT = 10.0
    item_map_cache = {}

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name: str, scope: str):
        return super().get(f"{scope}__{TargetModel.PRIMARY_KEY}", name)

    @classmethod
    def names(cls, scope: str):
        return super().names(f"{scope}__{TargetModel.PRIMARY_KEY}")

    @classmethod
    def all(cls, scope: str):
        return super().all(f"{scope}__{TargetModel.PRIMARY_KEY}")

    @classmethod
    def packet(
        cls,
        target_name: str,
        packet_name: str,
        type: str = "TLM",
        scope: str = OPENC3_SCOPE,
    ):
        """@return [Hash] Packet hash or raises an exception"""
        if type not in cls.VALID_TYPES:
            raise RuntimeError(f"Unknown type {type} for {target_name} {packet_name}")

        # Assume it exists and just try to get it to avoid an extra call to Store.exist?
        json_data = Store.hget(f"{scope}__openc3{type.lower()}__{target_name}", packet_name)
        if not json_data:
            raise RuntimeError(f"Packet '{target_name} {packet_name}' does not exist")
        return json.loads(json_data)

    @classmethod
    def packets(cls, target_name: str, type: str = "TLM", scope: str = OPENC3_SCOPE):
        """@return [Array>Hash>] All packet hashes under the target_name"""
        if type not in cls.VALID_TYPES:
            raise RuntimeError(f"Unknown type {type} for {target_name}")
        if not cls.get(name=target_name, scope=scope):
            raise RuntimeError(f"Target '{target_name}' does not exist")

        result = []
        packets = Store.hgetall(f"{scope}__openc3{type.lower()}__{target_name}")
        for _, packet_json in packets.items():
            result.append(json.loads(packet_json))
        return result

    @classmethod
    def set_packet(
        cls,
        target_name: str,
        packet_name: str,
        packet: Any,
        type: str = "TLM",
        scope: str = OPENC3_SCOPE,
    ):
        if type not in cls.VALID_TYPES:
            raise RuntimeError(f"Unknown type {type} for {target_name} {packet_name}")

        try:
            Store.hset(
                f"{scope}__openc3{type.lower()}__{target_name}",
                packet_name,
                json.dumps(packet),
            )
        except (TypeError, RuntimeError) as error:
            Logger.error(f"Invalid text present in {target_name} {packet_name} {type.lower()} packet")
            raise error

    @classmethod
    def packet_item(
        cls,
        target_name: str,
        packet_name: str,
        item_name,
        type: str = "TLM",
        scope: str = OPENC3_SCOPE,
    ):
        """@return [Hash] Item hash or raises an exception"""
        packet = cls.packet(target_name, packet_name, type=type, scope=scope)
        found = None
        for item in packet["items"]:
            if item["name"] == item_name:
                found = item
                break
        if not found:
            raise RuntimeError(f"Item '{packet['target_name']} {packet['packet_name']} {item_name}' does not exist")
        return found

    # @return [Array<Hash>] Item hash array or raises an exception
    @classmethod
    def packet_items(
        cls,
        target_name: str,
        packet_name: str,
        items: Any,
        type: str = "TLM",
        scope: str = OPENC3_SCOPE,
    ):
        packet = cls.packet(target_name, packet_name, type=type, scope=scope)
        found = []
        for item in packet["items"]:
            if item["name"] in items:
                found.append(item)
        if len(found) != len(items):  # we didn't find them all
            found_items = [item["name"] for item in found]
            not_found = []
            for item in list(set(items) - set(found_items)):
                not_found.append(f"'{target_name} {packet_name} {item}'")
            # 'does not exist' not grammatically correct but we use it in every other exception
            raise RuntimeError(f"Item(s) {', '.join(not_found)} does not exist")
        return found

    # @return [List<String>] All the item names for every packet in a target
    @classmethod
    def all_item_names(cls, target_name: str, type: str = "TLM", scope: str = OPENC3_SCOPE):
        items = Store.zrange(f"{scope}__openc3tlm__{target_name}__allitems", 0, -1)
        if not items:
            items = cls.rebuild_target_allitems_list(target_name, type=type, scope=scope)
        return items

    @classmethod
    def rebuild_target_allitems_list(cls, target_name: str, type: str = "TLM", scope: str = OPENC3_SCOPE):
        for packet in cls.packets(target_name, scope=scope):
            for item in packet["items"]:
                cls.add_to_target_allitems_list(target_name, item["name"], scope=scope)
        return Store.zrange(f"{scope}__openc3tlm__{target_name}__allitems", 0, -1) # return the new sorted set to let redis do the sorting

    @classmethod
    def add_to_target_allitems_list(cls, target_name: str, item_name: str, scope: str = OPENC3_SCOPE):
        score = 0 # https://redis.io/docs/latest/develop/data-types/sorted-sets/#lexicographical-scores
        Store.zadd(f"{scope}__openc3tlm__{target_name}__allitems", score, item_name)

    # @return [Hash{String => Array<Array<String, String, String>>}]
    @classmethod
    def limits_groups(cls, scope: str = OPENC3_SCOPE):
        groups = Store.hgetall(f"{scope}__limits_groups")
        if groups:
            return {k.decode(): json.loads(v) for (k, v) in groups.items()}
        else:
            return {}

    @classmethod
    def get_item_to_packet_map(cls, target_name: str, scope: str = OPENC3_SCOPE):
        if target_name in TargetModel.item_map_cache:
            cache_time, item_map = TargetModel.item_map_cache[target_name]
            if (time.time() - cache_time) < TargetModel.ITEM_MAP_CACHE_TIMEOUT:
                return item_map
        item_map_key = f"{scope}__{target_name}__item_to_packet_map"
        target_name = target_name.upper()
        json_data = Store.get(item_map_key)
        if json_data:
            item_map = json.loads(json_data)
        else:
            item_map = cls.build_item_to_packet_map(target_name, scope=scope)
            Store.set(item_map_key, json.dumps(item_map))
        TargetModel.item_map_cache[target_name] = [time.time(), item_map]
        return item_map

    @classmethod
    def build_item_to_packet_map(cls, target_name: str, scope: str = OPENC3_SCOPE):
        item_map = {}
        for packet in cls.packets(target_name, scope=scope):
            items = packet["items"]
            for item in items:
                item_name = item["name"]
                if item_map.get(item_name) is None:
                    item_map[item_name] = []
                item_map[item_name].append(packet["packet_name"])
        return item_map


    @classmethod
    def increment_telemetry_count(cls, target_name: str, packet_name: str, count: int, scope: str = OPENC3_SCOPE):
        result = Store.hincrby(f"{scope}__TELEMETRYCNTS__{{{target_name}}}", packet_name, count)
        if isinstance(result, (bytes, bytearray)):
            return int(result)
        else:
            return result


    @classmethod
    def get_all_telemetry_counts(cls, target_name: str, scope: str = OPENC3_SCOPE):
        result = {}
        get_all = Store.hgetall(f"{scope}__TELEMETRYCNTS__{{{target_name}}}")
        if get_all is dict:
            for key, value in get_all.items():
                result[key] = int(value)
            return result
        else:
            return get_all

    @classmethod
    def get_telemetry_count(cls, target_name: str, packet_name: str, scope: str = OPENC3_SCOPE):
        value = Store.hget(f"{scope}__TELEMETRYCNTS__{{{target_name}}}", packet_name)
        if value is None:
            return 0
        elif isinstance(value, (bytes, bytearray)):
            return int(value)
        else:
            return value

    @classmethod
    def get_telemetry_counts(cls, target_packets: list, scope: str = OPENC3_SCOPE):
        result = []
        with Store.instance().redis_pool.get() as redis:
            pipeline = redis.pipeline(transaction=False)
            for target_name, packet_name in target_packets:
                target_name = target_name.upper()
                packet_name = packet_name.upper()
                pipeline.hget(f"{scope}__TELEMETRYCNTS__{{{target_name}}}", packet_name)
            result = pipeline.execute()

        counts = []
        for count in result:
            if count is None:
                counts.append(0)
            else:
                counts.append(int(count))
        return counts

    @classmethod
    def increment_command_count(cls, target_name: str, packet_name: str , count: int, scope: str = OPENC3_SCOPE):
        result = Store.hincrby(f"{scope}__COMMANDCNTS__{{{target_name}}}", packet_name, count)
        if isinstance(result, (bytes, bytearray)):
            return int(result)
        else:
            return result

    @classmethod
    def get_all_command_counts(cls, target_name: str, scope: str = OPENC3_SCOPE):
        result = {}
        get_all = Store.hgetall(f"{scope}__COMMANDCNTS__{{{target_name}}}")
        if get_all is dict:
            for key, value in get_all.items():
                result[key] = int(value)
            return result
        else:
            return get_all

    @classmethod
    def get_command_count(cls, target_name: str, packet_name: str, scope: str = OPENC3_SCOPE):
        value = Store.hget(f"{scope}__COMMANDCNTS__{{{target_name}}}", packet_name)
        if value is None:
            return 0
        elif isinstance(value, (bytes, bytearray)):
            return int(value)
        else:
            return value

    @classmethod
    def get_command_counts(cls, target_packets: list, scope: str = OPENC3_SCOPE):
        result = []
        with Store.instance().redis_pool.get() as redis:
            pipeline = redis.pipeline(transaction=False)
            for target_name, packet_name in target_packets:
                target_name = target_name.upper()
                packet_name = packet_name.upper()
                pipeline.hget(f"{scope}__COMMANDCNTS__{{{target_name}}}", packet_name)
            result = pipeline.execute()

        counts = []
        for count in result:
            if count is None:
                counts.append(0)
            else:
                counts.append(int(count))
        return counts


    # Most of these parameters are unused but they must match the Ruby implementation
    # so we can call TargetModel.get_model which calls Model.get_model which does
    #   json_data = cls.get(name, scope)
    #   return cls.from_json(json_data, scope)
    # cls.from_json calls cls(**json_data) which is the constructor which takes
    # all the keyword arguments from the json_data which were set during installation
    # by the Ruby code
    def __init__(
        self,
        name: str,
        folder_name=None,
        requires=[],
        ignored_parameters=[],
        ignored_items=[],
        limits_groups=[],
        cmd_tlm_files=[],
        cmd_unique_id_mode=False,
        tlm_unique_id_mode=False,
        id=None,
        updated_at=None,
        plugin=None,
        cmd_buffer_depth=5,
        cmd_log_cycle_time=600,
        cmd_log_cycle_size=50_000_000,
        cmd_log_retain_time=None,
        cmd_decom_log_cycle_time=600,
        cmd_decom_log_cycle_size=50_000_000,
        cmd_decom_log_retain_time=None,
        tlm_buffer_depth=60,
        tlm_log_cycle_time=600,
        tlm_log_cycle_size=50_000_000,
        tlm_log_retain_time=None,
        tlm_decom_log_cycle_time=600,
        tlm_decom_log_cycle_size=50_000_000,
        tlm_decom_log_retain_time=None,
        reduced_minute_log_retain_time=None,
        reduced_hour_log_retain_time=None,
        reduced_day_log_retain_time=None,
        cleanup_poll_time=600,
        needs_dependencies=False,
        target_microservices={"REDUCER": [[]]},
        reducer_disable=False,
        reducer_max_cpu_utilization=30.0,
        disable_erb=None,
        shard=0,
        scope: str = OPENC3_SCOPE,
    ):
        super().__init__(
            f"{scope}__{self.PRIMARY_KEY}",
            name=name,
            updated_at=updated_at,
            plugin=plugin,
            scope=scope,
        )

    def update_store_telemetry(self, packet_hash, clear_old=True):
        for target_name, packets in packet_hash.items():
            if clear_old:
                Store.delete(f"{self.scope}__openc3tlm__{target_name}")
                Store.delete(f"{self.scope}__openc3tlm__{target_name}__allitems")
            for packet_name, packet in packets.items():
                Logger.debug(f"Configuring tlm packet= {target_name} {packet_name}")
                try:
                    Store.hset(f"{self.scope}__openc3tlm__{target_name}", packet_name, json.dumps(packet.as_json()))
                except Exception as e:
                    Logger.error(f"Invalid text present in {target_name} {packet_name} tlm packet")
                    raise e
                json_hash = {}
                for item in packet.sorted_items:
                    json_hash[item.name] = None
                    TargetModel.add_to_target_allitems_list(target_name, item.name, scope=self.scope)
                # Use Store.hset directly instead of CvtModel.set to avoid circular dependency
                Store.hset(
                    f"{self.scope}__tlm__{packet.target_name}",
                    packet.packet_name,
                    json.dumps(json_hash, cls=JsonEncoder),
                )

    def update_store_commands(self, packet_hash, clear_old=True):
        for target_name, packets in packet_hash.items():
            if clear_old:
                Store.delete(f"{self.scope}__openc3cmd__{target_name}")
            for packet_name, packet in packets.items():
                Logger.debug(f"Configuring cmd packet= {target_name} {packet_name}")
                try:
                    Store.hset(f"{self.scope}__openc3cmd__{target_name}", packet_name, json.dumps(packet.as_json()))
                except Exception as e:
                    Logger.error(f"Invalid text present in {target_name} {packet_name} cmd packet")
                    raise e

    def update_store_item_map(self):
        # Create item_map
        item_map_key = f"{self.scope}__{self.name}__item_to_packet_map"
        item_map = TargetModel.build_item_to_packet_map(self.name, scope=self.scope)
        Store.set(item_map_key, json.dumps(item_map))
        TargetModel.item_map_cache[self.name] = [time.time(), item_map]

    def dynamic_update(self, packets, cmd_or_tlm="TELEMETRY", filename="dynamic_tlm.txt"):
        # Build hash of targets/packets
        packet_hash = {}
        for packet in packets:
            target_name = packet.target_name.upper()
            if not packet_hash.get(target_name, None):
                packet_hash[target_name] = {}
            packet_name = packet.packet_name.upper()
            packet_hash[target_name][packet_name] = packet

        # Update Redis
        if cmd_or_tlm == "TELEMETRY":
            self.update_store_telemetry(packet_hash, clear_old=False)
            self.update_store_item_map()
        else:
            self.update_store_commands(packet_hash, clear_old=False)

        # Build dynamic file for cmd_tlm
        configs = {}
        for packet in packets:
            target_name = packet.target_name.upper()
            if not configs.get(target_name, None):
                configs[target_name] = ""
            config = configs[target_name]
            config += packet.to_config(cmd_or_tlm)
            config += "\n"
        for target_name, config in configs.items():
            bucket_key = f"{self.scope}/targets_modified/{target_name}/cmd_tlm/{filename}"
            client = Bucket.getClient()
            client.put_object(
                # Use targets_modified to save modifications
                # This keeps the original target clean (read-only)
                bucket=OPENC3_CONFIG_BUCKET,
                key=bucket_key,
                body=config,
            )

        # Inform microservices of new topics
        # Need to tell loggers to log, and decom to decom
        # We do this for no downtime
        raw_topics = []
        decom_topics = []
        for packet in packets:
            if cmd_or_tlm == "TELEMETRY":
                raw_topics.append(f"{self.scope}__TELEMETRY__{{{self.name}}}__{packet.packet_name.upper()}")
                decom_topics.append(f"{self.scope}__DECOM__{{{self.name}}}__{packet.packet_name.upper()}")
            else:
                raw_topics.append(f"{self.scope}__COMMAND__{{{self.name}}}__{packet.packet_name.upper()}")
                decom_topics.append(f"{self.scope}__DECOMCMD__{{{self.name}}}__{packet.packet_name.upper()}")
        if cmd_or_tlm == "TELEMETRY":
            Topic.write_topic(
                f"MICROSERVICE__{self.scope}__PACKETLOG__{self.name}",
                {"command": "ADD_TOPICS", "topics": json.dumps(raw_topics, cls=JsonEncoder)},
            )
            self.add_topics_to_microservice(f"{self.scope}__PACKETLOG__{self.name}", raw_topics)
            Topic.write_topic(
                f"MICROSERVICE__{self.scope}__DECOMLOG__{self.name}",
                {"command": "ADD_TOPICS", "topics": json.dumps(decom_topics, cls=JsonEncoder)},
            )
            self.add_topics_to_microservice(f"{self.scope}__DECOMLOG__{self.name}", decom_topics)
            Topic.write_topic(
                f"MICROSERVICE__{self.scope}__DECOM__{self.name}",
                {"command": "ADD_TOPICS", "topics": json.dumps(raw_topics, cls=JsonEncoder)},
            )
            self.add_topics_to_microservice(f"{self.scope}__DECOM__{self.name}", raw_topics)
        else:
            Topic.write_topic(
                f"MICROSERVICE__{self.scope}__COMMANDLOG__{self.name}",
                {"command": "ADD_TOPICS", "topics": json.dumps(raw_topics, cls=JsonEncoder)},
            )
            self.add_topics_to_microservice(f"{self.scope}__COMMANDLOG__{self.name}", raw_topics)
            Topic.write_topic(
                f"MICROSERVICE__{self.scope}__DECOMCMDLOG__{self.name}",
                {"command": "ADD_TOPICS", "topics": json.dumps(decom_topics, cls=JsonEncoder)},
            )
            self.add_topics_to_microservice(f"{self.scope}__DECOMCMDLOG__{self.name}", decom_topics)

    def add_topics_to_microservice(self, microservice_name, topics):
        model = MicroserviceModel.get_model(name=microservice_name, scope=self.scope)
        model.topics.extend(topics)
        model.topics = list(set(model.topics))
        model.ignore_changes = True  # Don't restart the microservice right now
        model.update()
